'''External API for rejester task system.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''
from __future__ import absolute_import, division
import abc
import logging
import os
import pkg_resources
import random
import psutil
import socket
import time
import traceback

from rejester._registry import Registry, nice_identifier
from rejester.exceptions import ProgrammerError, LockError, \
    LostLease, EnvironmentError, NoSuchWorkSpecError, NoSuchWorkUnitError

logger = logging.getLogger(__name__)

BUNDLES = 'bundles'
NICE_LEVELS = 'NICE_LEVELS'
WORK_SPECS = 'WORK_SPECS'
WORK_UNITS_ = 'WORK_UNITS_'
_BLOCKED = '_BLOCKED'
_BLOCKS = '_BLOCKS'
_DEPENDS = '_DEPENDS'
_FINISHED = '_FINISHED'
_FAILED = '_FAILED'
WORKER_CHILDREN_ = 'WORKER_CHILDREN_'
WORKER_STATE_ = 'WORKER_STATE_'
WORKER_OBSERVED_MODE = 'WORKER_OBSERVED_MODE'
ACTIVE_LEASES = 'ACTIVE_LEASES'


## work unit states
AVAILABLE = 1
BLOCKED = 2
PENDING = 3
FINISHED = 4
FAILED = 5

WORK_UNIT_STATUS_BY_NAME = {
    'AVAILABLE': AVAILABLE,
    'BLOCKED': BLOCKED,
    'PENDING': PENDING,
    'FINISHED': FINISHED,
    'FAILED': FAILED,
}

WORK_UNIT_STATUS_NAMES_BY_NUMBER = {
    v:k for k,v in WORK_UNIT_STATUS_BY_NAME.iteritems()
}


def build_task_master(config):
    tm_factory_module_name = config.get('task_master_module')
    if tm_factory_module_name is not None:
        import importlib
        tm_factory_module = importlib.import_module(tm_factory_module_name)
        tm_factory_clasname = config.get('task_master_class')
        assert tm_factory_clasname, 'task_master_module set but not task_master_class'
        _task_master_factory = getattr(tm_factory_module, tm_factory_clasname)
        return _task_master_factory(config)
    # by default use implementation later in this file
    return TaskMaster(config)


class Worker(object):
    '''Process that runs rejester jobs.

    Running a worker involves three steps: calling :meth:`register` to
    get a :attr:`worker_id` and record our presence in the data store;
    calling :meth:`run` to actually do work; and calling :meth:`unregister`
    on clean exit.  The :meth:`run` method should periodically call
    :meth:`heartbeat` to update its state and get the current run mode.

    ..automethod:: __init__

    '''

    __metaclass__ = abc.ABCMeta

    def __init__(self, config, task_master=None):
        '''Create a new worker.

        :param dict config: Configuration for the worker, typically
          the contents of the ``rejester`` block in the global config

        '''
        #: Configuration for the worker
        self.config = config
        #: Task interface to talk to the data store
        self.task_master = task_master
        if self.task_master is None:
            self.task_master = build_task_master(self.config)
        #: Worker ID, only valid after :meth:`register`
        self.worker_id = None
        #: Parent worker ID
        self.parent = None
        #: Required maximum time between :meth:`heartbeat`
        self.lifetime = 300 * 20  # 100 minutes

    def environment(self):
        '''Get raw data about this worker.

        This is recorded in the :meth:`heartbeat` info, and can be
        retrieved by :meth:`TaskMaster.get_heartbeat`.  The dictionary
        includes keys ``worker_id``, ``host``, ``fqdn``, ``version``,
        ``working_set``, and ``memory``.

        '''
        hostname = socket.gethostname()
        aliases = ()
        ipaddrs = ()

        # This sequence isn't 100% reliable.  We might try a socket()
        # sequence like RedisBase._ipaddress(), or just decide that
        # socket.fqdn() and/or socket.gethostname() is good enough.
        try:
            ip = socket.gethostbyname(hostname)
        except socket.herror:
            # If you're here, then $(hostname) doesn't resolve.
            ip = None

        try:
            if ip is not None:
                hostname, aliases, ipaddrs = socket.gethostbyaddr(ip)
        except socket.herror:
            # If you're here, then $(hostname) resolves, but the IP
            # address that results in doesn't reverse-resolve.  This
            # has been observed on OSX at least.
            ipaddrs = (ip,)

        env = dict(
            worker_id=self.worker_id,
            parent=self.parent,
            hostname=hostname,
            aliases=tuple(aliases),
            ipaddrs=tuple(ipaddrs),
            fqdn=socket.getfqdn(),
            version=pkg_resources.get_distribution("rejester").version, # pylint: disable=E1103
            working_set=[(dist.key, dist.version) for dist in pkg_resources.WorkingSet()], # pylint: disable=E1103
            # config_hash=self.config['config_hash'],
            # config_json = self.config['config_json'],
            memory=psutil.virtual_memory(),
            pid=os.getpid(),
        )
        return env

    def register(self, parent=None):
        '''Record the availability of this worker and get a unique identifer.

        This sets :attr:`worker_id` and calls :meth:`heartbeat`.  This
        cannot be called multiple times without calling
        :meth:`unregister` in between.

        '''
        if self.worker_id:
            raise ProgrammerError('Worker.register cannot be called again without first calling unregister; it is not idempotent')
        self.parent = parent
        self.worker_id = nice_identifier()
        self.task_master.worker_id = self.worker_id
        self.heartbeat()
        return self.worker_id

    def unregister(self):
        '''Remove this worker from the list of available workers.

        This requires the worker to already have been :meth:`register()`.

        ''' 
        self.task_master.worker_unregister(self.worker_id)
        self.task_master.worker_id = None
        self.worker_id = None

    def heartbeat(self):
        '''Record the current worker state in the registry.

        This records the worker's current mode, plus the contents of
        :meth:`environment`, in the data store for inspection by others.

        :returns mode: Current mode, as :meth:`TaskMaster.get_mode`

        '''
        mode = self.task_master.get_mode()
        self.task_master.worker_heartbeat(self.worker_id, mode,
                                          self.lifetime, self.environment(),
                                          parent=self.parent)
        return mode

    @abc.abstractmethod
    def run(self):
        '''Run some number of jobs.

        :meth:`register` must have already been called.  This is
        expected to get jobs using :meth:`TaskMaster.get_work` with
        this worker's :attr:`worker_id`.  Depending on the semantics
        of the actual implementing class this may run one job, run
        jobs as long as the worker's mode is :attr:`~TaskMaster.RUN`,
        or any other combination.

        '''
        return


class WorkUnit(object):
    '''A single unit of work being executed.

    These are created by the rejester system; the standard worker
    system will pass objects of this type to the named run function.
    If some code calls:

    .. code-block:: python

        task_master.update_bundle({ 'name': 'work_spec_name', ... },
                                  { 'key': data, ... })

    Then when the work unit is executed, this object will have the
    provided :attr:`work_spec_name`, :attr:`key`, and :attr:`data`,
    with remaining values being provided by the system.

    In the standard worker system, the worker will call :meth:`finish`
    or :meth:`fail` as appropriate.  The run function should examine
    :attr:`spec`, :attr:`key`, and :attr:`data` to figure out what to
    do.

    .. automethod:: __init__

    '''
    def __init__(self, registry, work_spec_name, key, data, worker_id=None,
                 expires=None, default_lifetime=900):
        '''Create a new work unit runtime data.

        In most cases application code will not need to call this directly,
        but should expect to be passed work units created by the system.

        :param registry: Mid-level Redis interface
        :type registry: :class:`rejester.Registry`
        :param str work_spec_name: Name of the work spec
        :param str key: Name of the work unit
        :param dict data: Data provided for the work unit
        :param str worker_id: Worker doing this work unit
        :param int expires: Latest time this work unit can still be running
        :param int default_lifetime: Time :meth:`update` adds by default

        '''
        if not worker_id:
            raise ProgrammerError('must specify a worker_id, not: %r' % worker_id)
        #: Worker doing this work unit
        self.worker_id = worker_id
        #: Mid-level Redis interface
        self.registry = registry
        #: Name of the work spec
        self.work_spec_name = work_spec_name
        #: Name of the work unit
        self.key = key
        #: Data provided for the work unit
        self.data = data
        #: Has this work unit called :meth:`finish`?
        self.finished = False
        #: Has this work unit called :meth:`fail`?
        self.failed = False
        #: Time (as :func:`time.time`) when this work unit must finish
        self.expires = expires
        #: Time :meth:`update` adds by default
        self.default_lifetime = default_lifetime
        self._spec_cache = None  # storage for lazy getter property
        self._module_cache = None  # storage for lazy getter property

    def __repr__(self):
        return ('WorkUnit(work_spec_name={0.work_spec_name}, key={0.key})'
                .format(self))

    @property
    def spec(self):
        '''Actual work spec.

        This is retrieved from the database on first use, and in some cases
        a worker can be mildly more efficient if it avoids using this.

        '''
        if self._spec_cache is None:
            self._spec_cache = self.registry.get(WORK_SPECS, self.work_spec_name)
        return self._spec_cache

    @property
    def module(self):
        '''Python module to run the job.

        This is used by :func:`run` and the standard worker system.
        If the work spec contains keys ``module``, ``run_function``,
        and ``terminate_function``, then this contains the Python
        module object named as ``module``; otherwise this contains
        :const:`None`.

        '''
        if self._module_cache is None:
            funclist = filter(None, (self.spec.get('run_function'), self.spec.get('terminate_function')))
            if funclist:
                try:
                    self._module_cache = __import__(
                        self.spec['module'], globals(), (), funclist, -1)
                except Exception, exc:
                    logger.error('failed to load spec["module"] = %r',
                                 self.spec['module'], exc_info=True)
                    raise
        return self._module_cache

    def run(self):
        '''Actually runs the work unit.

        This is called by the standard worker system, generally
        once per work unit.  It requires the work spec to contain
        keys ``module``, ``run_function``, and ``terminate_function``.
        It looks up ``run_function`` in :attr:`module` and calls that
        function with :const:`self` as its only parameter.

        '''
        try:
            logger.info('running work unit {}'.format(self.key))
            run_function = getattr(self.module, self.spec['run_function'])
            ret_val = run_function(self)
            self.update()
            logger.info('completed work unit {}'.format(self.key))
            return ret_val
        except LostLease, exc:
            logger.warning('work unit {} timed out'.format(self.key))
            raise
        except Exception, exc:
            logger.error('work unit {} failed'.format(self.key),
                         exc_info=True)
            raise

    def terminate(self):
        '''Kills the work unit.

        This is called by the standard worker system, but only in
        response to an operating system signal.  If the job does setup
        such as creating a child process, its terminate function
        should kill that child process.  More specifically, this
        function requires the work spec to contain the keys
        ``module``, ``run_function``, and ``terminate_function``, and
        calls ``terminate_function`` in :attr:`module` containing
        :const:`self` as its only parameter.

        '''
        terminate_function_name = self.spec.get('terminate_function')
        if not terminate_function_name:
            logger.error('tried to terminate WorkUnit(%r) but no function name', self.key)
            return None
        terminate_function = getattr(self.module, self.spec['terminate_function'])
        if not terminate_function:
            logger.error('tried to terminate WorkUnit(%r) but no function %s in module %r', self.key, terminate_function_name, self.module.__name__)
            return None
        logger.info('calling terminate function for work unit {}'
                    .format(self.key))
        ret_val = terminate_function(self)
        self.update(lease_time=-10)
        return ret_val

    def update(self, lease_time=None):
        '''Refresh this task's expiration time.

        This tries to set the task's expiration time to the current
        time, plus `lease_time` seconds.  It requires the job to not
        already be complete.  If `lease_time` is negative, makes the
        job immediately be available for other workers to run.

        :param int lease_time: time to extend job lease beyond now
        :raises rejester.exceptions.LostLease: if the lease has already
          expired

        '''
        if self.finished:
            raise ProgrammerError('cannot .update() after .finish()')
        if self.failed:
            raise ProgrammerError('cannot .update() after .fail()')
        if lease_time is None:
            lease_time = self.default_lifetime
        with self.registry.lock(identifier=self.worker_id) as session:
            if self.finished:
                logger.debug('WorkUnit(%r) became finished while waiting for lock to update', self.key)
                return
            if self.failed:
                logger.debug('WorkUnit(%r) became failed while waiting for lock to update', self.key)
                return
            try:
                self.expires = time.time() + lease_time
                session.update(
                    WORK_UNITS_ + self.work_spec_name,
                    {self.key: self.data}, 
                    priorities={self.key: self.expires},
                    locks={self.key: self.worker_id})

            except EnvironmentError, exc:
                raise LostLease(exc)
                
    def finish(self):
        '''Move this work unit to a finished state.

        In the standard worker system, the worker calls this on the job's
        behalf when :meth:`run_function` returns successfully.  Does nothing
        if the work unit is already complete.

        '''
        if self.finished or self.failed:
            return
        with self.registry.lock(identifier=self.worker_id) as session:
            if self.finished or self.failed:
                return
            logger.debug('WorkUnit finish %r', self.key)
            session.move(
                WORK_UNITS_ + self.work_spec_name,
                WORK_UNITS_ + self.work_spec_name + _FINISHED,
                {self.key: self.data})
            blocks = session.get(WORK_UNITS_ + self.work_spec_name + _BLOCKS,
                                 self.key)
            if blocks is not None:
                for block in blocks:
                    spec = block[0]
                    unit = block[1]
                    hard = block[2]
                    depends = session.get(WORK_UNITS_ + spec + _DEPENDS, unit)
                    if depends is None: continue
                    depends.remove([self.work_spec_name, self.key])
                    if len(depends) == 0:
                        session.popmany(WORK_UNITS_ + spec + _DEPENDS, unit)
                        unitdef = session.get(WORK_UNITS_ + spec + _BLOCKED,
                                              unit)
                        session.move(WORK_UNITS_ + spec + _BLOCKED,
                                     WORK_UNITS_ + spec,
                                     { unit: unitdef })
                    else:
                        session.set(WORK_UNITS_ + spec + _DEPENDS, unit,
                                    depends)
        self.finished = True

    def fail(self, exc=None):
        '''Move this work unit to a failed state.

        In the standard worker system, the worker calls this on the job's
        behalf when :meth:`run_function` ends with any exception:

        .. code-block:: python

            try:
                work_unit.run()
                work_unit.finish()
            except Exception, e:
                work_unit.fail(e)

        A ``traceback`` property is recorded with a formatted version
        of `exc`, if any.  Does nothing if the work unit is already
        complete.

        :param exc: Exception that caused the failure, or :const:`None`

        '''
        if self.failed or self.finished:
            return
        self.data['traceback'] = exc and traceback.format_exc(exc) or exc
        #self.data['worker_state'] = self.worker_state
        with self.registry.lock(identifier=self.worker_id) as session:
            if self.finished or self.failed:
                return
            session.move(
                WORK_UNITS_ + self.work_spec_name,
                WORK_UNITS_ + self.work_spec_name + _FAILED,
                {self.key: self.data})
            blocks = session.get(WORK_UNITS_ + self.work_spec_name + _BLOCKS,
                                 self.key)
            if blocks is not None:
                for block in blocks:
                    spec = block[0]
                    unit = block[1]
                    hard = block[2]
                    if hard:
                        session.popmany(WORK_UNITS_ + spec + _DEPENDS, unit)
                        unitdef = session.get(WORK_UNITS_ + spec + _BLOCKED,
                                              unit)
                        if unitdef is not None:
                            session.move(WORK_UNITS_ + spec + _BLOCKED,
                                         WORK_UNITS_ + spec + _FAILED,
                                         { unit: unitdef })
                    else:
                        depends = session.get(WORK_UNITS_ + spec + _DEPENDS, unit)
                        if depends is None: continue
                        depends.remove([self.work_spec_name, self.key])
                        if len(depends) == 0:
                            session.popmany(WORK_UNITS_ + spec + _DEPENDS, unit)
                            unitdef = session.get(WORK_UNITS_ + spec + _BLOCKED,
                                                  unit)
                            if unitdef is not None:
                                session.move(WORK_UNITS_ + spec + _BLOCKED,
                                             WORK_UNITS_ + spec,
                                             { unit: unitdef })
                        else:
                            session.set(WORK_UNITS_ + spec + _DEPENDS, unit,
                                        depends)
        self.failed = True


class TaskMaster(object):
    '''Control object for the rejester task queue.

    The task queue consists of a series of *work specs*, which include
    configuration information, and each work spec has some number of
    *work units* attached to it.  Both the work specs and work units
    are defined as (non-empty) dictionaries.  The work spec must have
    keys ``name`` and ``min_gb``, but any other properties are permitted.
    Conventionally ``desc`` contains a description of the job and
    ``config`` contains the top-level global configuration.

    There are three ways to use :class:`TaskMaster`:

    1. Create work specs and work units with :meth:`update_bundle`.
       Directly call :meth:`get_work` to get work units back.  Based
       on the information stored in the work spec and work unit
       dictionaries, do the work manually, and call
       :meth:`WorkUnit.finish` or :meth:`WorkUnit.fail` as
       appropriate.

    2. Create work specs and work units with :meth:`update_bundle`.
       The work spec must contain three keys, ``module`` naming a Python
       module, and ``run_function`` and ``terminate_function`` each
       naming functions of a single parameter in that module.  Directly
       call :meth:`get_work` to get work units back, then call their
       :meth:`WorkUnit.run` function to execute them.  See the basic
       example in :meth:`WorkUnit.fail`.

    3. Create work specs and work units with :meth:`update_bundle`,
       including the Python information.  Use one of the standard worker
       implementations in :mod:`rejester.workers` to actually run the
       job.

    Most applications will use the third option, the standard worker
    system.  In all three cases populating and executing jobs can
    happen on different systems, or on multiple systems in parallel.
    To use the standard worker system, create a Python module::

        def rejester_run(work_unit):
            # Does the actual work for `work_unit`.
            # Must be a top-level function in the module.
            # The work unit will succeed if this returns normally,
            # and will fail if this raises an exception.
            pass

        def rejester_terminate(work_unit):
            # Called only if a signal terminates the worker.
            # This usually does nothing, but could kill a known subprocess.
            pass

    The work spec would look something like::

        work_spec = {
            'name': 'rejester_sample',
            'desc': 'A sample rejester job.',
            'min_gb': 1,
            'config': yakonfig.get_global_config(),
            'module': 'name.of.the.module.from.above',
            'run_function': 'rejester_run',
            'terminate_function': 'rejester_terminate',
        }

    The work units can be any non-empty dictionaries that are
    meaningful to the run function.  :class:`WorkUnit` objects are
    created with their :attr:`~WorkUnit.key` and :attr:`~WorkUnit.data`
    fields set to individual keys and values from the parameter to
    :meth:`update_bundle`.

    A work unit can be in one of five states.  It is *available* if it
    has been added to the queue but nobody is working on it.  It is
    *pending* if somebody is currently working on it.  When they
    finish, it will become either *finished* or *failed*.  If
    dependencies are added between tasks using `add_dependent_task()`,
    a task can also be *blocked*.

    The general flow for a rejester application is to :meth:`set_mode`
    to :attr:`IDLE`, then add work units using
    :meth:`update_bundle()`, and :meth:`set_mode` to :attr:`RUN`.
    :meth:`get_work` will return work units until all have been
    consumed.  :meth:`set_mode` to :attr:`TERMINATE` will instruct
    workers to shut down.

    This object keeps very little state locally and can safely be used
    concurrently, including from multiple systems.  Correspondingly,
    any settings here, including :meth:`set_mode`, are persistent even
    beyond the end of the current process.

    .. automethod:: __init__

    '''

    def __init__(self, config):
        '''Create a new task master.

        This is a lightweight object, and it is safe to have multiple
        objects concurrently accessing the same rejester system, even
        on multiple machines.

        :param dict config: Configuration for the task master, generally
          the contents of the ``rejester`` block in the global configuration

        '''
        config['app_name'] = 'rejester'
        #: Configuration for the task master
        self.config = config
        #: Mid-level Redis interface
        self.registry = Registry(config)
        #: Worker ID, if this is tied to a worker
        self.worker_id = None
        #: Amount of time workers get between :meth:`WorkUnit.update` calls
        self.default_lifetime = config.get('default_lifetime', 900)
        #: Override available-memory checks
        self.enough_memory = config.get('enough_memory', False)

    #: Mode constant instructing workers to do work
    RUN = 'RUN'
    #: Mode constant instructing workers to not start new work
    IDLE = 'IDLE'
    #: Mode constant instructing workers to shut down
    TERMINATE = 'TERMINATE'

    # states for a work unit
    AVAILABLE = AVAILABLE
    BLOCKED = BLOCKED
    PENDING = PENDING
    FINISHED = FINISHED
    FAILED = FAILED

    def set_mode(self, mode):
        '''Set the global mode of the rejester system.

        This must be one of the constants :attr:`TERMINATE`,
        :attr:`RUN`, or :attr:`IDLE`.  :attr:`TERMINATE` instructs any
        running workers to do an orderly shutdown, completing current
        jobs then exiting.  :attr:`IDLE` instructs workers to stay
        running but not start new jobs.  :attr:`RUN` tells workers to
        do actual work.

        :param str mode: new rejester mode
        :raise rejester.exceptions.ProgrammerError: on invalid `mode`

        '''
        if mode not in [self.TERMINATE, self.RUN, self.IDLE]:
            raise ProgrammerError('mode=%r is not recognized' % mode)
        with self.registry.lock(identifier=self.worker_id) as session:
            session.set('modes', 'mode', mode)
        logger.info('set mode to %s', mode)

    def get_mode(self):
        '''Get the global mode of the rejester system.

        :return: rejester mode, :attr:`IDLE` if unset

        '''
        return self.registry.get('modes', 'mode') or self.IDLE

    def idle_all_workers(self):
        '''Set the global mode to :attr:`IDLE` and wait for workers to stop.

        This can wait arbitrarily long before returning.  The worst
        case in "normal" usage involves waiting five minutes for a
        "lost" job to expire; a well-behaved but very-long-running job
        can extend its own lease further, and this function will not
        return until that job finishes (if ever).

        .. deprecated:: 0.4.5
            There isn't an obvious use case for this function, and its
            "maybe wait forever for something out of my control" nature
            makes it hard to use in real code.  Polling all of the work
            specs and their :meth:`num_pending` in application code if
            you really needed this operation would have the same
            semantics and database load.

        '''
        self.set_mode(self.IDLE)
        while 1:
            num_pending = dict()
            for work_spec_name in self.registry.pull(NICE_LEVELS).keys():
                num_pending[work_spec_name] = self.num_pending(work_spec_name)
            if sum(num_pending.values()) == 0:
                break
            logger.warn('waiting for pending work_units: %r', num_pending)
            time.sleep(1)

    def mode_counts(self):
        '''Get the number of workers in each mode.

        This returns a dictionary where the keys are mode constants
        and the values are a simple integer count of the number of
        workers in that mode.

        '''
        modes = {self.RUN: 0, self.IDLE: 0, self.TERMINATE: 0}
        for worker_id, mode in self.workers().items():
            modes[mode] += 1
        return modes

    def workers(self, alive=True):
        '''Get a listing of all workers.

        This returns a dictionary mapping worker ID to the mode
        constant for their last observed mode.

        :param bool alive: if true (default), only include workers
          that have called :meth:`Worker.heartbeat` sufficiently recently

        '''
        return self.registry.filter(
            WORKER_OBSERVED_MODE,
            priority_min=alive and time.time() or '-inf')

    def get_heartbeat(self, worker_id):
        '''Get the last known state of some worker.

        If the worker never existed, or the worker's lifetime has
        passed without it heartbeating, this will return an empty
        dictionary.

        :param str worker_id: worker ID
        :return: dictionary of worker state, or empty dictionary
        :see: :meth:`Worker.heartbeat`

        '''
        return self.registry.pull(WORKER_STATE_ + worker_id)

    def dump(self):
        '''Print the entire contents of this to debug log messages.

        This is really only intended for debugging.  It could produce
        a lot of data.

        '''
        with self.registry.lock(identifier=self.worker_id) as session:
            for work_spec_name in self.registry.pull(NICE_LEVELS).iterkeys():
                def scan(sfx):
                    v = self.registry.pull(WORK_UNITS_ + work_spec_name + sfx)
                    if v is None: return []
                    return v.keys()
                for key in scan(''):
                    logger.debug('spec {} unit {} available or pending'
                                 .format(work_spec_name, key))
                for key in scan(_BLOCKED):
                    blocked_on = self.registry.get(
                        WORK_UNITS_ + work_spec_name + _DEPENDS, key)
                    logger.debug('spec {} unit {} blocked on {!r}'
                                 .format(work_spec_name, key, blocked_on))
                for key in scan(_FINISHED):
                    logger.debug('spec {} unit {} finished'
                                 .format(work_spec_name, key))
                for key in scan(_FAILED):
                    logger.debug('spec {} unit {} failed'
                                 .format(work_spec_name, key))


    @classmethod
    def validate_work_spec(cls, work_spec):
        '''Check that `work_spec` is valid.

        It must at the very minimum contain a ``name`` and ``min_gb``.

        :raise rejester.exceptions.ProgrammerError: if it isn't valid

        '''
        if 'name' not in work_spec:
            raise ProgrammerError('work_spec lacks "name"')
        if 'min_gb' not in work_spec or \
                not isinstance(work_spec['min_gb'], (float, int)):
            raise ProgrammerError('work_spec["min_gb"] must be a number')

    def num_available(self, work_spec_name):
        '''Get the number of available work units for some work spec.

        These are work units that could be returned by :meth:`get_work`:
        they are not complete, not currently executing, and not blocked
        on some other work unit.

        '''
        return self.registry.len(WORK_UNITS_ + work_spec_name, 
                                 priority_max=time.time())

    def num_pending(self, work_spec_name):
        '''Get the number of pending work units for some work spec.

        These are work units that some worker is currently working on
        (hopefully; it could include work units assigned to workers that
        died and that have not yet expired).

        '''
        return self.registry.len(WORK_UNITS_ + work_spec_name, 
                                 priority_min=time.time())

    def num_blocked(self, work_spec_name):
        '''Get the number of blocked work units for some work spec.

        These are work units that are the first parameter to
        :meth:`add_dependent_work_units` where the job they depend on
        has not yet completed.

        '''
        return self.registry.len(WORK_UNITS_ + work_spec_name + _BLOCKED)

    def num_finished(self, work_spec_name):
        '''Get the number of finished work units for some work spec.

        These have completed successfully with :meth:`WorkUnit.finish`.

        '''
        return self.registry.len(WORK_UNITS_ + work_spec_name + _FINISHED)

    def num_failed(self, work_spec_name):
        '''Get the number of failed work units for some work spec.

        These have completed unsuccessfully with :meth:`WorkUnit.fail`.

        '''
        return self.registry.len(WORK_UNITS_ + work_spec_name + _FAILED)

    def num_tasks(self, work_spec_name):
        '''Get the total number of work units for some work spec.'''
        return self.num_finished(work_spec_name) + \
               self.num_failed(work_spec_name) + \
               self.registry.len(WORK_UNITS_ + work_spec_name)

    def status(self, work_spec_name):
        '''Get a summary dictionary for some work spec.

        The keys are the strings :meth:`num_available`, :meth:`num_pending`,
        :meth:`num_blocked`, :meth:`num_finished`, :meth:`num_failed`,
        and :meth:`num_tasks`, and the values are the values returned
        from those functions.

        '''
        return dict(
            num_available=self.num_available(work_spec_name),
            num_pending=self.num_pending(work_spec_name),
            num_blocked=self.num_blocked(work_spec_name),
            num_finished=self.num_finished(work_spec_name),
            num_failed=self.num_failed(work_spec_name),
            num_tasks=self.num_tasks(work_spec_name),
            )

    def list_work_specs(self, limit=None, start=None):
        '''Get the list of [(work spec name, work spec), ...]

        The keys are the work spec names; the values are the actual
        work spec definitions.

        return [(spec name, spec), ...], next start value

        '''
        return self.registry.pull(WORK_SPECS).items(), None

    def iter_work_specs(self, limit=None, start=None):
        '''
        yield work spec dicts
        '''
        count = 0
        ws_list, start = self.list_work_specs(limit, start)
        while True:
            for name_spec in ws_list:
                yield name_spec[1]
                count += 1
                if (limit is not None) and (count >= limit):
                    break
            if not start:
                break
            if limit is not None:
                limit -= count
            ws_list, start = self.list_work_specs(limit, start)

    def get_work_spec(self, work_spec_name):
        '''Get the dictionary defining some work spec.'''
        with self.registry.lock(identifier=self.worker_id) as session:
            return session.get(WORK_SPECS, work_spec_name)

    def get_work_units(self, work_spec_name, work_unit_keys=None, state=None, limit=None, start=None):
        '''
        options={work_unit_keys=None, state=None, limit=None, start=None}
        options['state'] accepts either a single state or a list/tuple of states

        returns [(wu key, wu data), ...]
        '''
        if work_unit_keys is not None:
            raise NotImplementedError("get_work_units(by work_unit_keys)")
        if start is None:
            start = 0
        if state is not None:
            if state == AVAILABLE:
                return self.list_available_work_units(work_spec_name, start=start, limit=limit).items()
            if state == PENDING:
                return self.list_pending_work_units(work_spec_name, start=start, limit=limit).items()
            if state == BLOCKED:
                return self.list_blocked_work_units(work_spec_name, start=start, limit=limit).items()
            if state == FINISHED:
                return self.list_finished_work_units(work_spec_name, start=start, limit=limit).items()
            if state == FAILED:
                return self.list_failed_work_units(work_spec_name, start=start, limit=limit).items()
            raise ProgrammerError("unknown state {!r}".format(state))
        return self.list_work_units(work_spec_name, start=start, limit=limit)

    def list_work_units(self, work_spec_name, start=0, limit=None):
        """Get a dictionary of work units for some work spec.

        The dictionary is from work unit name to work unit definiton.
        Only work units that have not been completed ("available" or
        "pending" work units) are included.

        """
        return self.registry.filter(WORK_UNITS_ + work_spec_name,
                                    start=start, limit=limit)

    def list_available_work_units(self, work_spec_name, start=0, limit=None):
        """Get a dictionary of available work units for some work spec.

        The dictionary is from work unit name to work unit definiton.
        Only work units that have not been started, or units that were
        started but did not complete in a timely fashion, are
        included.

        """
        return self.registry.filter(WORK_UNITS_ + work_spec_name,
                                    priority_max=time.time(),
                                    start=start, limit=limit)

    def list_pending_work_units(self, work_spec_name, start=0, limit=None):
        """Get a dictionary of in-progress work units for some work spec.

        The dictionary is from work unit name to work unit definiton.
        Units listed here should be worked on by some worker.

        """
        return self.registry.filter(WORK_UNITS_ + work_spec_name,
                                    priority_min=time.time(),
                                    start=start, limit=limit)

    def list_blocked_work_units(self, work_spec_name, start=0, limit=None):
        """Get a dictionary of blocked work units for some work spec.

        The dictionary is from work unit name to work unit definiton.
        Work units included in this list are blocked because they were
        listed as the first work unit in
        :func:`add_dependent_work_units`, and the work unit(s) they
        depend on have not completed yet.  This function does not tell
        why work units are blocked, it merely returns the fact that
        they are.

        """
        return self.registry.filter(WORK_UNITS_ + work_spec_name + _BLOCKED,
                                    start=start, limit=limit)

    def list_finished_work_units(self, work_spec_name, start=0, limit=None):
        """Get a dictionary of finished work units for some work spec.

        The dictionary is from work unit name to work unit definiton.
        Only work units that have been successfully completed are
        included.

        """
        return self.registry.filter(WORK_UNITS_ + work_spec_name + _FINISHED,
                                    start=start, limit=limit)

    def list_failed_work_units(self, work_spec_name, start=0, limit=None):
        """Get a dictionary of failed work units for some work spec.

        The dictionary is from work unit name to work unit definiton.
        Only work units that have completed unsuccessfully are included.

        """
        return self.registry.filter(WORK_UNITS_ + work_spec_name + _FAILED,
                                    start=start, limit=limit)

    def _remove_some_work_units(self, work_spec_name, work_unit_names,
                                suffix='', priority_min='-inf',
                                priority_max='+inf'):
        '''Remove some units from somewhere.'''
        now = time.time()
        if work_unit_names is None:
            count = 0
            while True:
                with self.registry.lock(identifier=self.worker_id) as session:
                    names = session.filter(
                        WORK_UNITS_ + work_spec_name + suffix,
                        priority_min=priority_min, priority_max=priority_max,
                        limit=1000)
                    if not names: break
                    count += session.popmany(
                        WORK_UNITS_ + work_spec_name + suffix, *names)
        else:
            # TODO: This needs to honor priority_min/priority_max,
            # otherwise it gets the wrong answer for "available"/
            # "pending" (it will get both states).
            with self.registry.lock(identifier=self.worker_id) as session:
                count = session.popmany(WORK_UNITS_ + work_spec_name + suffix,
                                        *work_unit_names)
        return count

    def del_work_units(self, work_spec_name, work_unit_keys=None,
                       state=None, all=False):
        '''Delete work units from a work spec.

        The parameters are considered in order as follows:

        * If `all` is :const:`True`, then all work units in
          `work_spec_name` are deleted; otherwise
        * If `state` is not :const:`None`, then all work units
          in the named state are deleted; otherwise
        * If `work_unit_keys` are specified, then those specific
          work units are deleted; otherwise
        * Nothing is deleted.

        :param str work_spec_name: name of the work spec
        :param list work_unit_keys: if not :const:`None`, only delete
          these specific keys
        :param str state: only delete work units in this state
        :param bool all: if true, delete all work units
        :return: number of work units deleted

        '''
        count = 0
        if (state is None) or (state == AVAILABLE):
            count += self.remove_available_work_units(work_spec_name, work_unit_keys)
        if (state is None) or (state == PENDING):
            count += self.remove_pending_work_units(work_spec_name, work_unit_keys)
        if (state is None) or (state == BLOCKED):
            count += self.remove_blocked_work_units(work_spec_name, work_unit_keys)
        if (state is None) or (state == FAILED):
            count += self.remove_failed_work_units(work_spec_name, work_unit_keys)
        if (state is None) or (state == FINISHED):
            count += self.remove_finished_work_units(work_spec_name, work_unit_keys)
        return count

    def remove_available_work_units(self, work_spec_name, work_unit_names):
        '''Remove some work units in the available queue.

        If `work_unit_names` is :const:`None` (which must be passed
        explicitly), all available work units in `work_spec_name` are
        removed; otherwise only the specific named work units will be.

        :param str work_spec_name: name of the work spec
        :param list work_unit_names: names of the work units, or
          :const:`None` for all in `work_spec_name`
        :return: number of work units removed

        '''
        return self._remove_some_work_units(
            work_spec_name, work_unit_names, priority_max=time.time())

    def remove_pending_work_units(self, work_spec_name, work_unit_names):
        '''Remove some work units in the pending list.

        If `work_unit_names` is :const:`None` (which must be passed
        explicitly), all pending work units in `work_spec_name` are
        removed; otherwise only the specific named work units will be.

        Note that this function has the potential to confuse workers
        if they are actually working on the work units in question.  If
        you have ensured that the workers are dead and you would be
        otherwise waiting for the leases to expire before calling
        :meth:`remove_available_work_units`, then this is a useful
        shortcut.

        :param str work_spec_name: name of the work spec
        :param list work_unit_names: names of the work units, or
          :const:`None` for all in `work_spec_name`
        :return: number of work units removed

        '''
        return self._remove_some_work_units(
            work_spec_name, work_unit_names, priority_min=time.time())

    def remove_blocked_work_units(self, work_spec_name, work_unit_names):
        '''Remove some work units in the blocked list.

        If `work_unit_names` is :const:`None` (which must be passed
        explicitly), all pending work units in `work_spec_name` are
        removed; otherwise only the specific named work units will be.

        Note that none of the "remove" functions will restart blocked
        work units, so if you have called
        e.g. :meth:`remove_available_work_units` for a predecessor
        job, you may need to also call this method for its successor.

        :param str work_spec_name: name of the work spec
        :param list work_unit_names: names of the work units, or
          :const:`None` for all in `work_spec_name`
        :return: number of work units removed

        '''
        return self._remove_some_work_units(
            work_spec_name, work_unit_names, suffix=_BLOCKED)

    def remove_failed_work_units(self, work_spec_name, work_unit_names):
        '''Remove some failed work units.

        If `work_unit_names` is :const:`None` (which must be passed
        explicitly), all failed work units in `work_spec_name` are
        removed; otherwise only the specific named work units will be.

        Also consider :meth:`retry` to move failed work units back into
        the available queue.

        :param str work_spec_name: name of the work spec
        :param list work_unit_names: names of the work units, or
          :const:`None` for all in `work_spec_name`
        :return: number of work units removed

        '''
        return self._remove_some_work_units(
            work_spec_name, work_unit_names, suffix=_FAILED)

    def remove_finished_work_units(self, work_spec_name, work_unit_names):
        '''Remove some finished work units.

        If `work_unit_names` is :const:`None` (which must be passed
        explicitly), all finished work units in `work_spec_name` are
        removed; otherwise only the specific named work units will be.

        :param str work_spec_name: name of the work spec
        :param list work_unit_names: names of the work units, or
          :const:`None` for all in `work_spec_name`
        :return: number of work units removed

        '''
        return self._remove_some_work_units(
            work_spec_name, work_unit_names, suffix=_FINISHED)

    def get_work_unit_status(self, work_spec_name, work_unit_key):
        '''Get a high-level status for some work unit.

        The return value is a dictionary.  The only required key is
        ``status``, which could be any of:

        ``missing``
          The work unit does not exist anywhere
        ``available``
          The work unit is available for new workers; additional
          keys include ``expiration`` (may be 0)
        ``pending``
          The work unit is being worked on; additional keys include
          ``expiration`` and ``worker_id`` (usually)
        ``blocked``
          The work unit is waiting for some other work units to finish;
          additional keys include ``depends_on``
        ``finished``
          The work unit has completed
        ``failed``
          The work unit failed; additional keys include ``traceback``

        :param str work_spec_name: name of the work spec
        :param str work_unit_name: name of the work unit
        :return: dictionary description of summary status
        '''
        with self.registry.lock(identifier=self.worker_id) as session:
            # In the available list?
            (unit,priority) = session.get(WORK_UNITS_ + work_spec_name,
                                          work_unit_key, include_priority=True)
            if unit:
                result = {}
                if priority < time.time():
                    result['status'] = 'available'
                else:
                    result['status'] = 'pending'
                result['expiration'] = priority
                # ...is anyone working on it?
                worker = session.get(WORK_UNITS_ + work_spec_name + "_locks",
                                     work_unit_key)
                if worker:
                    result['worker_id'] = worker
                return result

            # In the finished list?
            unit = session.get(WORK_UNITS_ + work_spec_name + _FINISHED,
                               work_unit_key)
            if unit:
                return { 'status': 'finished' }

            # In the failed list?
            unit = session.get(WORK_UNITS_ + work_spec_name + _FAILED,
                               work_unit_key)
            if unit:
                result = { 'status': 'failed' }
                if 'traceback' in unit:
                    result['traceback'] = unit['traceback']
                return result

            # In the blocked list?
            unit = session.get(WORK_UNITS_ + work_spec_name + _BLOCKED,
                               work_unit_key)
            if unit:
                # This should always have *something*, right?
                deps = session.get(WORK_UNITS_ + work_spec_name + _DEPENDS,
                                   work_unit_key, default=[])
                result = { 'status': 'blocked',
                           'depends_on': deps }
                return result

        return { 'status': 'missing' }

    def inspect_work_unit(self, work_spec_name, work_unit_key):
        '''Get the data for some work unit.

        Returns the data for that work unit, or `None` if it really
        can't be found.

        :param str work_spec_name: name of the work spec
        :param str work_unit_key: name of the work unit
        :return: definition of the work unit, or `None`
        '''
        with self.registry.lock(identifier=self.worker_id) as session:
            work_unit_data = session.get(
                WORK_UNITS_ + work_spec_name, work_unit_key)
            if not work_unit_data:
                work_unit_data = session.get(
                    WORK_UNITS_ + work_spec_name + _BLOCKED, work_unit_key)
            if not work_unit_data:
                work_unit_data = session.get(
                    WORK_UNITS_ + work_spec_name + _FINISHED, work_unit_key)
            if not work_unit_data:
                work_unit_data = session.get(
                    WORK_UNITS_ + work_spec_name + _FAILED, work_unit_key)
            return work_unit_data

    def _inspect_work_unit(self, session, work_spec_name, work_unit_key):
        return work_unit_data


    def reset_all(self, work_spec_name):
        '''Restart a work spec.

        This calls :meth:`idle_all_workers`, then moves all finished
        jobs back into the available queue.

        .. deprecated:: 0.4.5
            See :meth:`idle_all_workers` for problems with that method.
            This also ignores failed jobs and work unit dependencies.
            In practice, whatever generated a set of work units
            initially can recreate them easily enough.

        '''
        self.idle_all_workers()
        with self.registry.lock(identifier=self.worker_id) as session:
            session.move_all(WORK_UNITS_ + work_spec_name + _FINISHED,
                             WORK_UNITS_ + work_spec_name)
            session.reset_priorities(WORK_UNITS_ + work_spec_name, 0)

    def update_bundle(self, work_spec, work_units, nice=0):
        '''Load a work spec and some work units into the task list.
        
        update the work_spec and work_units.  Overwrites any existing
        work spec with the same ``work_spec['name']`` and similarly
        overwrites any WorkUnit with the same ``work_unit.key``

        :param dict work_spec: Work spec dictionary
        :param work_units: Keys are used as :attr:`WorkUnit.key`, values
          are used as :attr:`WorkUnit.data`
        :type work_units: dict of dict
        :param int nice: Niceness of `work_spec`, higher value is lower
          priority

        '''
        self.validate_work_spec(work_spec)
        
        work_spec_name = work_spec['name']
        with self.registry.lock(identifier=self.worker_id) as session:
            session.update(NICE_LEVELS, {work_spec_name: nice})
            session.update(WORK_SPECS,  {work_spec_name: work_spec})
            
            ## redis chokes on lua scripts with 
            work_units = work_units.items()
            window_size = 10**4
            for i in xrange(0, len(work_units), window_size):
                self.registry.re_acquire_lock()
                session.update(WORK_UNITS_ + work_spec_name, 
                               dict(work_units[i: i + window_size]))

    def set_work_spec(self, work_spec):
        '''
        work_spec is a dict()
        work_spec['name'] is used as work_spec_name in other API calls
        work_spec['nice'] is used for prioritization, if set.
        '''
        self.validate_work_spec(work_spec)

        nice = int(work_spec.get('nice', 0))
        work_spec_name = work_spec['name']
        with self.registry.lock(identifier=self.worker_id) as session:
            session.update(NICE_LEVELS, {work_spec_name: nice})
            session.update(WORK_SPECS,  {work_spec_name: work_spec})

    def add_work_units(self, work_spec_name, work_unit_key_vals):
        '''
        work_unit_key_vals list of (work_unit_key, work_unit_data)
        '''
        with self.registry.lock(identifier=self.worker_id) as session:
            window_size = 10**4
            for i in xrange(0, len(work_unit_key_vals), window_size):
                self.registry.re_acquire_lock()
                session.update(WORK_UNITS_ + work_spec_name,
                               dict(work_unit_key_vals[i: i + window_size]))

    def add_dependent_work_units(self, work_unit, depends_on, hard=True):
        """Add work units, where one prevents execution of the other.

        The two work units may be attached to different work specs,
        but both must be in this task master's namespace.  `work_unit`
        and `depends_on` are both tuples of (work spec name, work unit
        name, work unit dictionary).  The work specs must already
        exist; they may be created with :meth:`update_bundle` with
        an empty work unit dictionary.  If a work unit dictionary is
        provided with either work unit, then this defines that work
        unit, and any existing definition is replaced.  Either or both
        work unit dictionaries may be :const:`None`, in which case the
        work unit is not created if it does not already exist.  In
        this last case, the other work unit will be added if
        specified, but the dependency will not be added, and this
        function will return :const:`False`.  In all other cases, this
        dependency is added in addition to all existing dependencies
        on either or both work units, even if the work unit dictionary
        is replaced.

        `work_unit` will not be executed or reported as available via
        :meth:`get_work` until `depends_on` finishes execution.  If
        the `depends_on` task fails, then the `hard` parameter
        describes what happens: if `hard` is :const:`True` then
        `work_unit` will also fail, but if `hard` is :const:`False`
        then `work_unit` will be able to execute even if `depends_on`
        fails, it just must have completed some execution attempt.

        Calling this function with ``hard=True`` suggests an ordered
        sequence of tasks where the later task depends on the output
        of the earlier tasks.  Calling this function with
        ``hard=False`` suggests a cleanup task that must run after
        this task (and, likely, several others) are done, but doesn't
        specifically depend on its result being available.

        :param work_unit: "Later" work unit to execute
        :paramtype work_unit: tuple of (str,str,dict)
        :param depends_on: "Earlier" work unit to execute
        :paramtype depends_on: tuple of (str,str,dict)
        :param bool hard: if True, then `work_unit` automatically fails
          if `depends_on` fails
        :return: :const:`True`, unless one or both of the work units
          didn't exist and weren't specified, in which case, :const:`False`
        :raise rejester.exceptions.NoSuchWorkSpecError: if a work spec was
          named that doesn't exist

        """
        # There's no good, not-confusing terminology here.
        # I'll call work_unit "later" and depends_on "earlier"
        # consistently, because that at least makes the time flow
        # correct.
        later_spec, later_unit, later_unitdef = work_unit
        earlier_spec, earlier_unit, earlier_unitdef = depends_on
        with self.registry.lock(identifier=self.worker_id) as session:
            # Bail if either work spec doesn't already exist
            if session.get(WORK_SPECS, later_spec) is None:
                raise NoSuchWorkSpecError(later_spec)
            if session.get(WORK_SPECS, earlier_spec) is None:
                raise NoSuchWorkSpecError(earlier_spec)
            
            # Cause both work units to exist (if possible)
            # Note that if "earlier" is already finished, we may be
            # able to make "later" available immediately
            earlier_done = False
            earlier_successful = False
            if earlier_unitdef is not None:
                session.update(WORK_UNITS_ + earlier_spec,
                               { earlier_unit: earlier_unitdef })
            else:
                earlier_unitdef = session.get(
                    WORK_UNITS_ + earlier_spec, earlier_unit)
                if earlier_unitdef is None:
                    earlier_unitdef = session.get(
                        WORK_UNITS_ + earlier_spec + _BLOCKED, earlier_unit)
                if earlier_unitdef is None:
                    earlier_unitdef = session.get(
                        WORK_UNITS_ + earlier_spec + _FINISHED, earlier_unit)
                    if earlier_unitdef is not None:
                        earlier_done = True
                        earlier_successful = True
                if earlier_unitdef is None:
                    earlier_unitdef = session.get(
                        WORK_UNITS_ + earlier_spec + _FAILED, earlier_unit)
                    if earlier_unitdef is not None:
                        earlier_done = True

            later_failed = earlier_done and hard and not earlier_successful
            later_unblocked = ((earlier_done and not later_failed) or
                               (earlier_unitdef is None))
            if later_failed:
                later_destination = WORK_UNITS_ + later_spec + _FAILED
            elif later_unblocked:
                later_destination = WORK_UNITS_ + later_spec
            else:
                later_destination = WORK_UNITS_ + later_spec + _BLOCKED

            if later_unitdef is not None:
                for suffix in ['', _FINISHED, _FAILED, _BLOCKED]:
                    k = WORK_UNITS_ + later_spec + suffix
                    if k != later_destination:
                        session.popmany(k, later_unit)
                session.update(later_destination,
                               { later_unit: later_unitdef })
            elif earlier_unitdef is not None:
                later_unitdef = session.get(
                    WORK_UNITS_ + later_spec, later_unit)
                if later_unitdef is not None:
                    session.move(
                        WORK_UNITS_ + later_spec,
                        WORK_UNITS_ + later_spec + _BLOCKED,
                        { later_unit: later_unitdef })
                else:
                    later_unitdef = session.get(
                        WORK_UNITS_ + later_spec + _BLOCKED, later_unit)

            if later_unitdef is None or earlier_unitdef is None:
                return False

            # Now both units exist and are in the right place;
            # record the dependency
            blocks = session.get(WORK_UNITS_ + earlier_spec + _BLOCKS,
                                 earlier_unit)
            if blocks is None: blocks = []
            blocks.append([later_spec, later_unit, hard])
            session.set(WORK_UNITS_ + earlier_spec + _BLOCKS,
                        earlier_unit, blocks)

            depends = session.get(WORK_UNITS_ + later_spec + _DEPENDS,
                                  later_unit)
            if depends is None: depends = []
            depends.append([earlier_spec, earlier_unit])
            session.set(WORK_UNITS_ + later_spec + _DEPENDS,
                        later_unit, depends)

            return True

    def retry(self, work_spec_name, *work_unit_names):
        '''Move failed work unit(s) back into the "pending" queue.

        The work unit will be available to execute immediately.  If
        other tasks had depended on it, those dependencies will not
        be recreated.

        :param str work_spec_name: name of the (existing) work spec
        :param str work_unit_names: name(s) of the (failed) work unit(s)
        :raise rejester.NoSuchWorkSpecError: if `work_spec_name` is
          invalid
        :raise rejester.NoSuchWorkUnitError: if `work_spec_name` is
          valid but any of the `work_unit_names` are not a failed work unit
        :raise rejester.LockError: if the registry lock could not be
          obtained
        
        '''
        with self.registry.lock(identifier=self.worker_id) as session:
            # This sequence is safe even if this system dies
            units = {}
            for work_unit_name in work_unit_names:
                unit = session.get(WORK_UNITS_ + work_spec_name + _FAILED,
                                   work_unit_name)
                if unit is None:
                    spec = session.get(WORK_SPECS, work_spec_name)
                    if spec is None:
                        raise NoSuchWorkSpecError(work_spec_name)
                    else:
                        raise NoSuchWorkUnitError(work_unit_name)
                if 'traceback' in unit:
                    del unit['traceback']
                units[work_unit_name] = unit
            session.move(WORK_UNITS_ + work_spec_name + _FAILED,
                         WORK_UNITS_ + work_spec_name,
                         units,
                         priority=0)

    def nice(self, work_spec_name, nice):
        '''Change the priority of an existing work spec.'''
        with self.registry.lock(identifier=self.worker_id) as session:
            session.update(NICE_LEVELS, dict(work_spec_name=nice))

    def get_work(self, worker_id, available_gb=None, lease_time=None, work_spec_names=None):
        '''obtain a WorkUnit instance based on available memory for the
        worker process.  
        
        :param worker_id: unique identifier string for a worker to
          which a WorkUnit will be assigned, if available.
        :param available_gb: number of gigabytes of RAM available to
          this worker
        :param lease_time: how many seconds to lease a WorkUnit
        :param work_spec_names: limit to queue from one work_spec. NOT IMPLEMENTD. this implementation will return work from any work spec.

        '''

        if not isinstance(available_gb, (int, float)):
            raise ProgrammerError('must specify available_gb')

        if lease_time is None:
            lease_time = self.default_lifetime
        work_unit = None

        try: 
            with self.registry.lock(identifier=self.worker_id) as session:
                ## use the simple niceness algorithm described in
                ## http://en.wikipedia.org/wiki/Nice_(Unix)
                ## where each job gets a (20-niceness) share
                nice_levels = session.pull(NICE_LEVELS)
                for work_spec_name, nice in nice_levels.iteritems():
                    nice = min(19, nice)
                    nice = max(-19, nice)
                    nice = 20 - nice
                    nice_levels[work_spec_name] = nice

                while nice_levels:
                    total_nice = sum(nice_levels.values())
                    score = random.randrange(total_nice)
                    work_spec_name = None
                    total_score = 0
                    for wsn, nice in nice_levels.iteritems():
                        total_score += nice
                        if total_score > score:
                            work_spec_name = wsn
                            break
                    assert work_spec_name is not None
                    nice_levels.pop(work_spec_name)
                    
                    ## verify sufficient memory
                    work_spec = session.get(WORK_SPECS, work_spec_name)
                    if available_gb < work_spec['min_gb']:
                        if self.enough_memory:
                            logger.info('Not enough memory to run work '
                                        'spec %s (need %.1f GiB, have '
                                        '%.1f GiB) but running anyways',
                                        work_spec_name,
                                        work_spec['min_gb'],
                                        available_gb)
                        else:
                            logger.info('Not enough memory to run work '
                                        'spec %s (need %.1f GiB, have '
                                        '%.1f GiB)',
                                        work_spec_name,
                                        work_spec['min_gb'],
                                        available_gb)
                            continue

                    ## try to get a task
                    wu_expires = time.time() + lease_time
                    _work_unit = session.getitem_reset(
                        WORK_UNITS_ + work_spec_name,
                        priority_max=time.time(),
                        new_priority=wu_expires,
                        lock=worker_id,
                    )

                    if _work_unit:
                        logger.info('work unit %r', _work_unit)
                        work_unit = WorkUnit(
                            self.registry, work_spec_name,
                            _work_unit[0], _work_unit[1],
                            worker_id=worker_id,
                            expires=wu_expires,
                            default_lifetime=self.default_lifetime,
                        )
                        break

        except (LockError, EnvironmentError):
            logger.error('took to long to get work', exc_info=True)

        logger.debug('get_work %r', work_unit)
        return work_unit

    def get_assigned_work_unit(self, worker_id, work_spec_name,
                               work_unit_key):
        '''get a specific WorkUnit that has already been assigned to a
        particular worker_id
        '''
        with self.registry.lock(identifier=self.worker_id) as session:
            assigned_work_unit_key = session.get(
                WORK_UNITS_ + work_spec_name + '_locks', worker_id)
            if not assigned_work_unit_key == work_unit_key:
                # raise LostLease instead of EnvironmentError, so
                # users of TaskMaster can have a single type of
                # expected exception, rather than two
                raise LostLease(
                    'assigned_work_unit_key=%r != %r'
                    % (assigned_work_unit_key, work_unit_key))
            # could trap EnvironmentError and raise LostLease instead
            work_unit_data = session.get(WORK_UNITS_ + work_spec_name,
                                         work_unit_key)
            return WorkUnit(
                self.registry, work_spec_name,
                work_unit_key, work_unit_data,
                worker_id=worker_id,
                default_lifetime=self.default_lifetime,
            )

    def get_child_work_units(self, worker_id):
        '''Get work units assigned to a worker's children.

        Returns a dictionary mapping worker ID to :class:`WorkUnit`.
        If a child exists but is idle, that worker ID will map to
        :const:`None`.  The work unit may already be expired or
        assigned to a different worker; this will be reflected in
        the returned :class:`WorkUnit`.

        This may write back to the underlying data store to clean up
        stale children that have not unregistered themselves but
        no longer exist in any form.

        '''
        result = {}
        with self.registry.lock(identifier=worker_id) as session:
            all_children = session.pull(WORKER_CHILDREN_ + worker_id)
            # The data stored in Redis isn't actually conducive to
            # this specific query; we will need to scan each work spec
            # for each work unit
            work_specs = session.pull(WORK_SPECS)
            for child in all_children.iterkeys():
                work_spec_name = None
                for spec in work_specs.iterkeys():
                    work_unit_key = session.get(
                        WORK_UNITS_ + spec + '_locks', child)
                    if work_unit_key:
                        work_spec_name = spec
                        break
                if work_spec_name:
                    assigned = session.get(
                        WORK_UNITS_ + work_spec_name + '_locks',
                        work_unit_key)
                    (data, expires) = session.get(
                        WORK_UNITS_ + work_spec_name, work_unit_key,
                        include_priority=True)
                    result[child] = WorkUnit(
                        self.registry, work_spec_name, work_unit_key,
                        data, expires=expires, worker_id=assigned)
                else:
                    # The child isn't doing anything.  Does it still
                    # exist?
                    heartbeat = session.get(WORKER_OBSERVED_MODE, child)
                    if heartbeat:
                        result[child] = None
                    else:
                        session.popmany(WORKER_CHILDREN_ + worker_id,
                                        child)
        return result

    def worker_register(self, worker_id, mode=None, lifetime=6000,
                        environment=None, parent=None):
        # actually the same as heartbeat, just, "hello, I'm here"
        self.worker_heartbeat(worker_id, mode, lifetime, environment,
                              parent)

    def worker_heartbeat(self, worker_id, mode=None, lifetime=6000,
                         environment=None, parent=None):
        if environment is None:
            environment = {}
        with self.registry.lock(identifier=worker_id) as session:
            session.set(WORKER_OBSERVED_MODE, worker_id, mode,
                        priority=time.time() + lifetime)
            session.update(WORKER_STATE_ + worker_id, environment,
                           expire=lifetime)
            if parent:
                session.update(WORKER_CHILDREN_ + parent,
                               {worker_id: 1})

    def worker_unregister(self, worker_id, parent=None):
        with self.registry.lock(identifier=worker_id) as session:
            session.delete(WORKER_STATE_ + worker_id)
            session.delete(WORKER_CHILDREN_ + worker_id)
            session.popmany(WORKER_OBSERVED_MODE, worker_id)
            if parent:
                session.popmany(WORKER_CHILDREN_ + parent, worker_id)



"""Redis-based distributed priority queue.

Purpose
=======

``RejesterQueue`` provides a priority queue, where string values can
be inserted with some priority.  Items can be pulled from the queue in
priority order, and must be returned to the queue within some time
limit.  A queue item can also reserve other queue items, preventing
them from being checked out or reserved until the main item finishes.

Implementation Details
======================


This software is released under an MIT/X11 open source license.

Copyright 2014 Diffeo, Inc.
"""

from __future__ import absolute_import

import logging
import time

import redis

from rejester.exceptions import ItemInUseError, LostLease
from rejester._redis import RedisBase

logger = logging.getLogger(__name__)

class RejesterQueue(RedisBase):
    """A Redis-based priority queue.

    Queue items should generally be (short) strings.  Each item has
    some associated priority.  ``check_out_item()`` will return some
    item with the highest available priority.  The item must be
    ``renew_item()`` or ``return_item()`` within a specified
    expiration limit, or the item will be lost and may be given to
    another client.

    If you have a queue item checked out, you may also reserve other
    queue items.  This prevents other clients from checking out or
    reserving these items.  When you return the main queue item, or if
    your check out expires, reserved items will be released.

    Since the queue is based on Redis, there may be multiple queue
    clients on the same queue on different systems.  Each ``RejesterQueue``
    instance has a distinct worker ID, held in the ``worker_id`` property.
    You can manually specify it in the constructor or manually set it;
    otherwise, a unique ID will be obtained via the database on first use.

    """

    def __init__(self, config, name, worker_id=None):
        """Initialize the queue using a configuration object.

        ``config`` should be a dictionary with the following keys:

        ``registry_addresses``
          list of ``host:port`` for the Redis server(s)
        ``app_name``
          application name (set to "queue" if not provided)
        ``namespace``
          application invocation namespace name (should be unique per run)

        ``name`` is the name of the queue.  If ``worker_id`` is specified
        it sets the worker ID, otherwise the database will be used to get
        a unique one on first use.

        """
        config.setdefault('app_name', 'queue')
        super(RejesterQueue, self).__init__(config)
        self.name = name
        self._worker_id = worker_id

    def _conn(self):
        """Get a new redis connection."""
        return redis.StrictRedis(connection_pool=self.pool)

    # key name construction
    def _key_worker(self): return self._namespace('worker')
    def _key_for(self, k): return self._namespace(self.name + k)
    def _key_available(self): return self._key_for('_AVAILABLE')
    def _key_priorities(self): return self._key_for('_PRIORITIES')
    def _key_expiration(self): return self._key_for('_EXPIRATION')
    def _key_workers(self): return self._key_for('_WORKERS')
    def _key_reservations(self, i): return self._key_for('_RESERVATIONS_' + i)

    def dump_queue(self, *names):
        """Debug-log some of the queues.

        ``names`` may include any of "worker", "available", "priorities",
        "expiration", "workers", or "reservations_ITEM" filling in some
        specific item.

        """
        conn = redis.StrictRedis(connection_pool=self.pool)
        for name in names:
            if name == 'worker':
                logger.debug('last worker: ' + conn.get(self._key_worker()))
            elif name == 'available':
                logger.debug('available: ' +
                             str(conn.zrevrange(self._key_available(), 0, -1,
                                                withscores=True)))
            elif name == 'priorities':
                logger.debug('priorities: ' +
                             str(conn.hgetall(self._key_priorities())))
            elif name == 'expiration':
                logger.debug('expiration: ' +
                             str(conn.zrevrange(self._key_expiration(), 0, -1,
                                                withscores=True)))
            elif name == 'workers':
                logger.debug('workers: ' +
                             str(conn.hgetall(self._key_workers())))
            elif name.startswith('reservations_'):
                item = name[len('reservations_'):]
                logger.debug('reservations for ' + item + ': ' +
                             str(conn.smembers(self._key_reservations(item))))

    @property
    def worker_id(self):
        """A unique identifier for this queue instance and the items it owns."""
        if self._worker_id is not None: return self._worker_id
        return self._get_worker_id(self._conn())
    @worker_id.setter
    def worker_id(self, w):
        """Set the worker ID for this queue instance.

        If this is set to None, the property will be reset to a unique
        worker ID on next use.

        """
        self._worker_id = w

    def _get_worker_id(self, conn):
        """Get the worker ID, using a preestablished connection."""
        if self._worker_id is None:
            self._worker_id = conn.incr(self._key_worker())
        return self._worker_id

    def add_item(self, item, priority):
        """Add ``item`` to this queue.

        It will have the specified ``priority`` (highest priority runs
        first).  If it is already in the queue, fail if it is checked
        out or reserved, or change its priority to ``priority``
        otherwise.

        """
        conn = self._conn()
        self._run_expiration(conn)
        script = conn.register_script("""
        if (redis.call("hexists", KEYS[2], ARGV[1]) ~= 0) and
           not(redis.call("zscore", KEYS[1], ARGV[1]))
        then
            return -1
        end
        redis.call("zadd", KEYS[1], ARGV[2], ARGV[1])
        redis.call("hset", KEYS[2], ARGV[1], ARGV[2])
        return 0
        """)
        result = script(keys=[self._key_available(), self._key_priorities()],
                        args=[item, priority])
        if result == -1:
            raise ItemInUseError(item)
        return

    def check_out_item(self, expiration):
        """Get the highest-priority item out of this queue.

        Returns the item, or None if no items are available.  The item
        must be either ``return_item()`` or ``renew_item()`` before
        ``expiration`` seconds pass, or it will become available to
        future callers.  The item will be marked as being owned by
        ``worker_id``.

        """
        conn = redis.StrictRedis(connection_pool=self.pool)
        self._run_expiration(conn)
        expiration += time.time()
        script = conn.register_script("""
        local item = redis.call("zrevrange", KEYS[1], 0, 0)
        if #item == 0 then return nil end
        item = item[1]
        redis.call("zrem", KEYS[1], item)
        redis.call("zadd", KEYS[2], ARGV[1], item)
        redis.call("hset", KEYS[3], "i" .. item, "w" .. ARGV[2])
        redis.call("hset", KEYS[3], "w" .. ARGV[2], "i" .. item)
        return item
        """)
        result = script(keys=[self._key_available(), self._key_expiration(),
                              self._key_workers()],
                        args=[expiration, self._get_worker_id(conn)])
        return result

    def renew_item(self, item, expiration):
        """Update the expiration time for ``item``.

        The item will remain checked out for ``expiration`` seconds
        beyond the current time.  This queue instance must have
        already checked out ``item``, and this method can fail if
        ``item`` is already overdue.

        """
        conn = self._conn()
        self._run_expiration(conn)
        expiration += time.time()
        script = conn.register_script("""
        -- already expired?
        if redis.call("hget", KEYS[2], "i" .. ARGV[1]) ~= "w" .. ARGV[3]
        then return -1 end

        -- otherwise just update the expiration
        redis.call("zadd", KEYS[1], ARGV[2], ARGV[1])
        return 0
        """)
        result = script(keys=[self._key_expiration(), self._key_workers()],
                        args=[item, expiration, self._get_worker_id(conn)])
        if result == -1:
            raise LostLease(item)
        return

    def return_item(self, item, priority):
        """Complete work on an item from ``check_out_item()``.

        If this instance no longer owns ``item``, raise ``LostLease``.
        If ``priority`` is None, the item is removed from the queue;
        otherwise it is re-added with the specified priority.  Any
        locked items associated with this item are unlocked.

        """
        conn = redis.StrictRedis(connection_pool=self.pool)
        self._run_expiration(conn)
        script = conn.register_script("""
        -- expired?
        if redis.call("hget", KEYS[4], "i" .. ARGV[1]) ~= "w" .. ARGV[3]
        then return -1 end

        -- will no longer expire
        redis.call("zrem", KEYS[2], ARGV[1])
        -- update priority, readd to available list
        if ARGV[2] == "None"
        then
          redis.call("hdel", KEYS[3], ARGV[1])
        else
          redis.call("hset", KEYS[3], ARGV[1], ARGV[2])
          redis.call("zadd", KEYS[1], ARGV[2], ARGV[1])
        end
        -- release all reservations
        local reservations = redis.call("smembers", KEYS[5])
        for i = 1, #reservations do
          local item = reservations[i]
          local pri = redis.call("hget", KEYS[3], item)
          redis.call("zadd", KEYS[1], pri, item)
        end
        -- clear out workers
        redis.call("hdel", KEYS[4], "i" .. ARGV[1])
        redis.call("hdel", KEYS[4], "w" .. ARGV[3])
        
        return 0
        """)
        # work around python -> redis -> lua marshaling
        if priority is None: priority = "None"
        result = script(keys=[self._key_available(), self._key_expiration(),
                              self._key_priorities(), self._key_workers(),
                              self._key_reservations(item)],
                        args=[item, priority, self._get_worker_id(conn)])
        if result == -1: raise LostLease(item)
        return

    def reserve_items(self, parent_item, *items):
        """Reserve a set of items until a parent item is returned.

        Prevent ``check_out_item()`` from returning any of ``items``
        until ``parent_item`` is completed or times out.  For each
        item, if it is not already checked out or reserved by some
        other parent item, it is associated with ``parent_item``, and
        the reservation will be released when ``parent_item``
        completes or times out.  Returns a list that is a subset of
        ``items`` for which we could get the reservation.

        Raises ``LostLease`` if this queue instance no longer owns
        ``parent_item``.  If any of the items do not exist, they are
        silently ignored.

        """
        conn = redis.StrictRedis(connection_pool=self.pool)
        self._run_expiration(conn)
        script = conn.register_script("""
        -- expired?
        if redis.call("hget", KEYS[2], "i" .. ARGV[1]) ~= "w" .. ARGV[2]
        then return -1 end

        -- loop through each item
        local result = {}
        for i = 3, #ARGV do
          local item = ARGV[i]
          -- item must be available to reserve
          if redis.call("zscore", KEYS[1], item) then
            redis.call("zrem", KEYS[1], item)
            redis.call("sadd", KEYS[3], item)
            result[#result + 1] = item
          end
        end
        return result
        """)
        result = script(keys=[self._key_available(), self._key_workers(),
                              self._key_reservations(parent_item)],
                        args=([parent_item, self._get_worker_id(conn)] +
                              list(items)))
        if result == -1:
            raise LostLease(parent_item)
        return result

    def _run_expiration(self, conn):
        """Return any items that have expired."""
        # The logic here is sufficiently complicated, and we need
        # enough random keys (Redis documentation strongly encourages
        # not constructing key names in scripts) that we'll need to
        # do this in multiple steps.  This means that, when we do
        # go in and actually expire things, we need to first check
        # that they're still running.

        # Get, and clear out, the list of expiring items
        now = time.time()
        script = conn.register_script("""
        local result = redis.call("zrangebyscore", KEYS[1], 0, ARGV[1])
        redis.call("zremrangebyscore", KEYS[1], 0, ARGV[1])
        return result
        """)
        expiring = script(keys=[self._key_expiration()], args=[time.time()])
        
        # Manually expire each item one by one
        script = conn.register_script("""
        -- item may have fallen out of the worker list, if someone finished
        -- at just the very last possible moment (phew!)
        local wworker = redis.call("hget", KEYS[3], "i" .. ARGV[1])
        if not wworker then return end

        -- we want to return item, plus everything it's reserved
        local to_return = redis.call("smembers", KEYS[4])
        to_return[#to_return + 1] = ARGV[1]
        for i = 1, #to_return do
          local pri = redis.call("hget", KEYS[2], to_return[i])
          redis.call("zadd", KEYS[1], pri, to_return[i])
        end
        -- already removed from expiration list
        -- remove from worker list too
        redis.call("hdel", KEYS[3], "i" .. ARGV[1])
        redis.call("hdel", KEYS[3], wworker)
        """)
        for item in expiring:
            script(keys=[self._key_available(), self._key_priorities(),
                         self._key_workers(), self._key_reservations(item)],
                   args=[item])



'''
http://github.com/diffeo/rejester

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import, division
import logging
import os
import sys
import time

import pytest

from rejester import TaskMaster
from rejester._task_master import WORK_UNITS_, _FINISHED
from rejester.exceptions import LostLease, NoSuchWorkUnitError

logger = logging.getLogger(__name__)
pytest_plugins = 'rejester.tests.fixtures'

work_spec = dict(
    name = 'tbundle',
    desc = 'a test work bundle',
    min_gb = 8,
    config = dict(many='', params=''),
    module = 'tests.rejester.test_workers',
    run_function = 'work_program',
    terminate_function = 'work_program',
)

def test_task_master_basic_interface(task_master):
    work_units = dict(foo={}, bar={})
    task_master.update_bundle(work_spec, work_units)

    assert task_master.registry.pull(WORK_UNITS_ + work_spec['name'])

    ## check that we ccannot get a task with memory below min_gb
    assert task_master.get_work('fake_worker_id', available_gb=3) == None

    work_unit = task_master.get_work('fake_worker_id', available_gb=13)
    assert work_unit.key in work_units

    work_unit.data['status'] = 10
    work_unit.update()

    assert task_master.registry.pull(WORK_UNITS_ + work_spec['name'])[work_unit.key]['status'] == 10

    work_unit.finish()
    work_unit.finish()

    assert task_master.registry.pull(WORK_UNITS_ + work_spec['name'] + _FINISHED)[work_unit.key]['status'] == 10   

    assert 'status' in task_master.inspect_work_unit(work_spec['name'], work_unit.key)

def test_list_work_specs(task_master):
    # Initial state: nothing
    assert task_master.list_work_specs() == ([],None)
    
    work_units = dict(foo={ 'length': 3 }, foobar={ 'length': 6 })
    task_master.update_bundle(work_spec, work_units)

    specs, next = task_master.list_work_specs()
    specs = dict(specs)
    assert len(specs) == 1
    assert work_spec['name'] in specs
    assert specs[work_spec['name']]['desc'] == work_spec['desc']

def test_list_work_units(task_master):
    work_units = dict(foo={ 'length': 3 }, foobar={ 'length': 6 })
    task_master.update_bundle(work_spec, work_units)

    # Initial check: both work units are there
    u = task_master.list_work_units(work_spec['name'])
    for k,v in u.iteritems():
        assert k in work_units
        assert 'length' in v
        assert len(k) == v['length']
    for k in work_units.iterkeys(): assert k in u

    # Start one unit; should still be there
    work_unit = task_master.get_work('fake_worker_id', available_gb=13)
    assert work_unit.key in work_units
    u = task_master.list_work_units(work_spec['name'])
    assert work_unit.key in u
    for k in u.iterkeys(): assert k in work_units
    for k in work_units.iterkeys(): assert k in u

    # Finish that unit; should be gone, the other should be there
    work_unit.finish()
    u = task_master.list_work_units(work_spec['name'])
    assert work_unit.key not in u
    for k in u.iterkeys(): assert k in work_units
    for k in work_units.iterkeys(): assert k == work_unit.key or k in u

def test_list_work_units_start_limit(task_master):
    work_units = dict(foo={ 'length': 3 }, bar={ 'length': 6 })
    task_master.update_bundle(work_spec, work_units)

    u = task_master.list_work_units(work_spec['name'], start=0, limit=1)
    assert u == { 'bar': { 'length': 6 } }

    u = task_master.list_work_units(work_spec['name'], start=1, limit=1)
    assert u == { 'foo': { 'length': 3 } }

    u = task_master.list_work_units(work_spec['name'], start=2, limit=1)
    assert u == { }

def test_task_master_reset_all(task_master):
    work_units = dict(foo={}, bar={})
    task_master.update_bundle(work_spec, work_units)
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_pending(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == 2

    work_unit = task_master.get_work('fake_worker_id', available_gb=13)
    work_unit.data['status'] = 10
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == 1
    assert task_master.num_pending(work_spec['name']) == 1

    work_unit.update()
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == 1
    assert task_master.num_pending(work_spec['name']) == 1

    work_unit.finish()
    assert task_master.num_finished(work_spec['name']) == 1
    assert task_master.num_available(work_spec['name']) == 1
    assert task_master.num_pending(work_spec['name']) == 0

    task_master.reset_all(work_spec['name'])
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == 2
    assert task_master.num_pending(work_spec['name']) == 0

    assert len(task_master.registry.pull(WORK_UNITS_ + work_spec['name'])) == 2
    with task_master.registry.lock() as session:
        assert session.popitem(WORK_UNITS_ + work_spec['name'], priority_max=-1) is None

def test_task_master_retry(task_master):
    work_units = dict(foo={}, bar={})
    task_master.update_bundle(work_spec, work_units)
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_pending(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == 2

    work_unit = task_master.get_work('fake_worker_id', available_gb=13)
    wuname = work_unit.key
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == 1
    assert task_master.num_pending(work_spec['name']) == 1

    work_unit.fail(exc=Exception())
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_pending(work_spec['name']) == 0
    assert task_master.num_failed(work_spec['name']) == 1
    assert task_master.num_available(work_spec['name']) == 1

    task_master.retry(work_spec['name'], wuname)
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_pending(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == 2

    with pytest.raises(NoSuchWorkUnitError):
        task_master.retry(work_spec['name'], wuname)

    assert (sorted(task_master.list_work_units(work_spec['name'])) ==
            ['bar', 'foo'])

    work_unit = task_master.get_work('fake_worker_id', available_gb=13)
    assert 'traceback' not in work_unit.data
    work_unit.finish()
    work_unit = task_master.get_work('fake_worker_id', available_gb=13)
    assert 'traceback' not in work_unit.data
    work_unit.finish()
    work_unit = task_master.get_work('fake_worker_id', available_gb=13)
    assert work_unit is None

def test_task_master_lost_lease(task_master):
    '''test that waiting too long to renew a lease allows another worker
    to get the lease and leads to LostLease exception in the worker
    that waited too long.
    '''
    work_units = dict(task_key_42=dict(data='hello'))
    task_master.update_bundle(work_spec, work_units)

    ## lease a WorkUnit with very short lease_time
    work_unit1 = task_master.get_work('fake_worker_id1', available_gb=13, lease_time=1)
    time.sleep(2)
    work_unit2 = task_master.get_work('fake_worker_id2', available_gb=13, lease_time=10)

    assert work_unit1.key == work_unit2.key
    assert work_unit1.worker_id == 'fake_worker_id1'
    assert work_unit2.worker_id == 'fake_worker_id2'

    with pytest.raises(LostLease):
        work_unit1.update()


def test_worker_child(task_master):
    '''test the basic parent/child worker interface'''
    task_master.worker_register('child', mode=task_master.get_mode(),
                                parent='parent')
    try:
        assert task_master.get_child_work_units('parent') == {'child': None}
        assert task_master.get_child_work_units('child') == {}

        task_master.update_bundle(work_spec, {'k': {'kk': 'vv'}})
        wu = task_master.get_work('child', available_gb=13)
        assert wu is not None
        assert wu.key == 'k'

        cwus = task_master.get_child_work_units('parent')
        assert cwus.keys() == ['child']
        cwu = cwus['child']
        assert cwu.worker_id == 'child'
        assert cwu.work_spec_name == work_spec['name']
        assert cwu.key == wu.key
        assert cwu.expires == wu.expires

        assert task_master.get_child_work_units('child') == {}
    finally:
        task_master.worker_unregister('child', parent='parent')


def test_worker_child_expiry(task_master):
    '''test the parent/child interface when a job expires'''
    task_master.worker_register('child', mode=task_master.get_mode(),
                                parent='parent')
    try:
        assert task_master.get_child_work_units('parent') == {'child': None}
        assert task_master.get_child_work_units('child') == {}

        task_master.update_bundle(work_spec, {'k': {'kk': 'vv'}})
        wu = task_master.get_work('child', available_gb=13, lease_time=1)
        assert wu is not None
        assert wu.key == 'k'

        time.sleep(2)
        # Now the job is technically expired

        cwus = task_master.get_child_work_units('parent')
        assert cwus.keys() == ['child']
        cwu = cwus['child']
        assert cwu.worker_id == 'child'
        assert cwu.work_spec_name == work_spec['name']
        assert cwu.key == wu.key
        assert cwu.expires == wu.expires

        assert task_master.get_child_work_units('child') == {}
    finally:
        task_master.worker_unregister('child', parent='parent')


def test_worker_child_stolen(task_master):
    '''test the parent/child interface when a job expires'''
    task_master.worker_register('child', mode=task_master.get_mode(),
                                parent='parent')
    try:
        assert task_master.get_child_work_units('parent') == {'child': None}
        assert task_master.get_child_work_units('child') == {}

        task_master.update_bundle(work_spec, {'k': {'kk': 'vv'}})
        wu = task_master.get_work('child', available_gb=13, lease_time=1)
        assert wu is not None
        assert wu.key == 'k'

        time.sleep(2)
        wu = task_master.get_work('thief', available_gb=13)
        assert wu is not None
        assert wu.key == 'k'

        cwus = task_master.get_child_work_units('parent')
        assert cwus.keys() == ['child']
        cwu = cwus['child']
        assert cwu.worker_id == 'thief'
        assert cwu.work_spec_name == work_spec['name']
        assert cwu.key == wu.key
        assert cwu.expires == wu.expires

        assert task_master.get_child_work_units('child') == {}
    finally:
        task_master.worker_unregister('child', parent='parent')



"""py.test support for rejester

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

Purpose
=======

This module provides common py.test fixtures for various rejester
features.  They will clean up after themselves, taking care to not, for
instance, leak namespaces in the underlying redis database.


Implementation Notes
====================

You will need to tell pytest to load this module.  You can add to
either your test module or a shared ``conftest.py`` file

>>> pytest_plugins = 'rejester.tests.fixtures'

Your tests will then have fixtures like ``task_master`` and
``rejester_queue`` available to them, with automatic namespace
management.  Your test directory should also have a file named
``config_registry.yaml`` that looks like

    registry_addresses: ['redis.example.com:6379']

You can usually get the namespace string from the configuration of
one of the high-level objects.  If you must manually work with the
namespace string outside of the managed objects, you can call
``make_namespace_string()`` or depend on the ``namespace_string``
fixture.

"""

import getpass
import hashlib
import logging
import os
import re
import socket

import pytest
import yaml

from rejester import RejesterQueue, TaskMaster, Registry

def _then_delete_namespace(request, obj, registry=None):
    """Helper to delete a namespace on cleanup.

    :param request: py.test request fixture
    :param obj: object to return
    :param registry: object to call ``delete_namespace()``, use
      `obj` if `None`
    :return: obj

    """
    if registry is None: registry = obj
    def fin():
        # Avoid a spurious "no handlers could be found for logger"
        # complaint.  delete_namespace() has a (useful) debug log
        # statement; but when we get here, under the py.test
        # capturelog plugin, the root logger has already been
        # removed.
        handler = logging.NullHandler()
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        registry.delete_namespace()
        root_logger.removeHandler(handler)
    request.addfinalizer(fin)
    return obj

@pytest.fixture(scope='function')
def namespace_string2(request):
    """A second dynamically constructed namespace string.

    This is similar to `namespace_string`, but is guaranteed to be
    different.  Again, this is not managed on its own but is intended
    to be used by higher-level fixtures, and callers are responsible
    for ensuring that the namespace is cleaned up afterwards.

    """
    return make_namespace_string(request.node.name + '2')

@pytest.fixture(scope='function')
def _rejester_config(redis_address, namespace_string):
    """Basic Rejester configuration."""
    return {
        'app_name': 'rejester',
        'registry_addresses': [ redis_address ],
        'namespace': namespace_string,
    }

@pytest.fixture(scope='function')
def registry(request, _rejester_config, namespace_string):
    """A standalone rejester Registry.

    This is a lower-level object that can directly access the Redis
    database.  Most tests will not need to access this."""
    config = dict(_rejester_config)
    config['app_name'] = 'rejester_test'
    config['namespace'] = namespace_string
    registry = Registry(config)
    return _then_delete_namespace(request, registry)

@pytest.fixture(scope='function')
def registry2(registry):
    '''A second copy of :func:`registry`.'''
    return Registry(registry.config)

@pytest.fixture(scope='function')
def task_master(request, _rejester_config, namespace_string):
    """A rejester TaskMaster for loading or querying work.

    The TaskMaster will have a dynamically created namespace.
    When the test completes the namespace will be torn down.
    If you need the namespace name for other purposes, it will be
    in ``task_master.config['namespace']``.

    """
    config = dict(_rejester_config)
    config['namespace'] = namespace_string
    tm = TaskMaster(config)
    return _then_delete_namespace(request, tm, tm.registry)

@pytest.fixture(scope='function')
def task_master2(request, _rejester_config, namespace_string2):
    """A rejester TaskMaster on a second namespace.

    The TaskMaster will have a dynamically created namespace, distinct
    from task_master's.  When the test completes the namespace will be
    torn down.  If you need the namespace name for other purposes, it
    will be in ``task_master2.config['namespace']``.

    """
    config = dict(_rejester_config)
    config['namespace'] = namespace_string2
    tm = TaskMaster(config)
    return _then_delete_namespace(request, tm, tm.registry)

@pytest.fixture(scope='function')
def rejester_queue(request, _rejester_config, namespace_string):
    """A RejesterQueue that will die at the end of execution.

    The queue will will be named 'queue', and will have a dynamically
    created namespace.  When the test completes the namespace will be
    torn down.  If you need the namespace for other purposes, it will
    be in ``rejester_queue.config['namespace']``.

    """
    config = dict(_rejester_config)
    config['namespace'] = namespace_string
    q = RejesterQueue(config, 'queue')
    return _then_delete_namespace(request, q)

@pytest.fixture(scope='function')
def rejester_queue2(rejester_queue):
    """A copy of rejester_queue with a distinct worker ID"""
    return RejesterQueue(rejester_queue.config, rejester_queue.name)



'''Tests for the 'rejester' command-line tool.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
from cStringIO import StringIO
import errno
import logging
import json
import os
import signal
import subprocess
import sys
import time

import pytest
import yaml

import rejester
from rejester.run import Manager
from rejester._task_master import Worker
from rejester.tests.fixtures import task_master, _rejester_config
import yakonfig

logger = logging.getLogger(__name__)

def test_config_from_env(redis_address, monkeypatch, namespace_string):
    host, port = redis_address.split(':', 2)
    monkeypatch.setenv('REDIS_PORT_6379_TCP_ADDR', host)
    monkeypatch.setenv('REDIS_PORT_6379_TCP_PORT', port)
    with yakonfig.defaulted_config(
            [rejester],
            config={'rejester': {'namespace': namespace_string}}):
        pass  # should not throw a configuration error

@pytest.yield_fixture
def global_config(_rejester_config):
    with yakonfig.defaulted_config([rejester],
                                   config={'rejester': _rejester_config}):
        yield yakonfig.get_global_config()

@pytest.fixture
def manager(global_config, task_master):
    mgr = Manager()
    mgr._task_master = task_master
    mgr.stdout = StringIO()
    return mgr

class TrivialWorker(Worker):
    def run(self):
        self.work_unit.run()

@pytest.yield_fixture
def worker(_rejester_config):
    w = TrivialWorker(_rejester_config)
    w.register()
    yield w
    w.unregister()

def rejester_cb(work_unit):
    pass

@pytest.fixture
def work_spec():
    return dict(
        name = 'tbundle',
        desc = 'a test work bundle',
        min_gb = 0.01,
        config = dict(many=' ', params=''),
        module = 'rejester.tests.test_cli',
        run_function = 'rejester_cb',
        terminate_function = 'rejester_cb',
    )

@pytest.fixture
def work_units(global_config):
    num_units = 11
    return { 'key-{}'.format(x): { 'sleep': 0.2, 'config': global_config }
             for x in xrange(num_units) }

@pytest.fixture
def work_spec_path(work_spec, tmpdir):
    tmpf = str(tmpdir.join("work_spec.yaml"))
    with open(tmpf, 'w') as f:
        f.write(yaml.dump(work_spec))
    return tmpf

@pytest.fixture
def work_units_path(work_units, tmpdir):
    tmpf = str(tmpdir.join('work_units.yaml'))
    with open(tmpf, 'w') as f:
        for k, v in work_units.iteritems():
            f.write(json.dumps({k: v}) + '\n')
    return tmpf

@pytest.fixture
def loaded(task_master, work_spec, work_units):
    task_master.update_bundle(work_spec, work_units)

@pytest.fixture
def worked(task_master, worker, loaded):
    '''3 finished, 2 failed, 1 pending, rest available'''
    for n in xrange(3):
        unit = task_master.get_work(worker.worker_id, available_gb=16)
        unit.finish()
    for n in xrange(2):
        unit = task_master.get_work(worker.worker_id, available_gb=16)
        unit.fail()
    task_master.get_work(worker.worker_id, available_gb=16)

def test_load_args(manager, work_spec_path, work_units_path):
    '''various tests for invalid arguments to "load"'''
    with pytest.raises(SystemExit):
        manager.runcmd('load', [])
    with pytest.raises(SystemExit):
        manager.runcmd('load', ['-w', work_spec_path])
    with pytest.raises(SystemExit):
        manager.runcmd('load', ['-u', work_units_path])
    with pytest.raises(SystemExit):
        manager.runcmd('load', ['-w', os.path.basename(work_spec_path),
                                '-u', work_units_path])
    with pytest.raises(SystemExit):
        manager.runcmd('load', ['-w', work_spec_path,
                                '-u', os.path.basename(work_units_path)])

def test_load(manager, work_spec_path, work_units_path, work_spec, work_units):
    manager.runcmd('load', ['-w', work_spec_path, '-u', work_units_path])

    tm = manager.task_master
    spec = tm.get_work_spec(work_spec['name'])
    assert spec['desc'] == work_spec['desc']
    units = tm.list_work_units(work_spec['name'])
    assert len(units) == len(work_units)
    assert sorted(units.keys()) == sorted(work_units.keys())

def test_delete(manager, loaded, work_spec, namespace_string):
    spec = manager.task_master.get_work_spec(work_spec['name'])
    assert spec['desc'] == work_spec['desc']

    manager.runcmd('delete', ['-y'])
    assert (manager.stdout.getvalue() ==
            'deleting namespace \'' + namespace_string + '\'\n')

    spec = manager.task_master.get_work_spec(work_spec['name'])
    assert spec is None

def test_work_specs_empty(manager):
    manager.runcmd('work_specs', [])
    assert manager.stdout.getvalue() == ''

def test_work_specs_loaded(manager, loaded, work_spec):
    manager.runcmd('work_specs', [])
    assert manager.stdout.getvalue() == work_spec['name'] + '\n'

def test_work_spec_bad(manager, loaded, tmpdir):
    with pytest.raises(SystemExit):
        manager.runcmd('work_spec', [])
    with pytest.raises(SystemExit):
        manager.runcmd('work_spec', ['-w', str(tmpdir.join('missing'))])

def test_work_spec_by_name(manager, loaded, work_spec):
    manager.runcmd('work_spec', ['-W', work_spec['name'], '--json'])
    spec = json.loads(manager.stdout.getvalue())
    assert spec == work_spec

def test_work_spec_by_file(manager, loaded, work_spec, work_spec_path):
    manager.runcmd('work_spec', ['-w', work_spec_path, '--json'])
    spec = json.loads(manager.stdout.getvalue())
    assert spec == work_spec

def test_status_bad(manager, loaded, tmpdir):
    with pytest.raises(SystemExit):
        manager.runcmd('status', [])
    with pytest.raises(SystemExit):
        manager.runcmd('status', ['-w', str(tmpdir.join('missing'))])

def test_status_by_name(manager, loaded, work_spec, work_units):
    manager.runcmd('status', ['-W', work_spec['name']])
    status = json.loads(manager.stdout.getvalue())
    ref = {
        'num_available': len(work_units),
        'num_blocked': 0,
        'num_failed': 0,
        'num_finished': 0,
        'num_pending': 0,
        'num_tasks': len(work_units),
    }
    assert status == ref

def test_status_by_file(manager, loaded, work_spec_path, work_units):
    manager.runcmd('status', ['-w', work_spec_path])
    status = json.loads(manager.stdout.getvalue())
    ref = {
        'num_available': len(work_units),
        'num_blocked': 0,
        'num_failed': 0,
        'num_finished': 0,
        'num_pending': 0,
        'num_tasks': len(work_units),
    }
    assert status == ref

def test_status_not_loaded(manager, work_spec):
    manager.runcmd('status', ['-W', work_spec['name']])
    status = json.loads(manager.stdout.getvalue())
    ref = {
        'num_available': 0,
        'num_blocked': 0,
        'num_failed': 0,
        'num_finished': 0,
        'num_pending': 0,
        'num_tasks': 0,
    }
    assert status == ref

def test_status_with_work(manager, worked, work_spec, work_units):
    manager.runcmd('status', ['-W', work_spec['name']])
    status = json.loads(manager.stdout.getvalue())
    ref = {
        'num_available': len(work_units)-6,
        'num_blocked': 0,
        'num_failed': 2,
        'num_finished': 3,
        'num_pending': 1,
        'num_tasks': len(work_units),
    }
    assert status == ref

def test_summary(manager, loaded, work_spec):
    manager.runcmd('summary', [])
    assert (manager.stdout.getvalue() ==
            'Work spec               Avail  Pending  Blocked'
            '   Failed Finished    Total\n'
            '==================== ======== ======== ========'
            ' ======== ======== ========\n'
            'tbundle                    11        0        0'
            '        0        0       11\n')

def test_summary_worked(manager, worked, work_spec):
    manager.runcmd('summary', [])
    assert (manager.stdout.getvalue() ==
            'Work spec               Avail  Pending  Blocked'
            '   Failed Finished    Total\n'
            '==================== ======== ======== ========'
            ' ======== ======== ========\n'
            'tbundle                     5        1        0'
            '        2        3       11\n')

def test_work_units_names(manager, loaded, work_spec, work_units):
    manager.runcmd('work_units', ['-W', work_spec['name']])
    response = manager.stdout.getvalue()
    names = response.strip().split('\n')
    assert sorted(names) == sorted(work_units.keys())

def test_work_units_details(manager, loaded, work_spec, work_units):
    manager.runcmd('work_units', ['-W', work_spec['name'], '--details'])
    response = manager.stdout.getvalue()
    work_units_json = list()
    work_unit_str = ''
    capturing = False
    for line in response.strip().split('\n'):
        if line.startswith("u'"):
            line = '{' + line[1:].replace("'", '"')
            capturing = True
        if capturing:
            work_unit_str += line
        if line.startswith("}"):
            capturing = False
            work_unit_str += '}'
            work_units_json.append(json.loads(work_unit_str))
            work_unit_str = ''
    found = set()
    for work_unit in work_units_json:
        key = work_unit.keys()[0]
        found.add(key)
    assert sorted(found) == sorted(work_units.keys())

def test_failed_trivial(manager, loaded, work_spec):
    manager.runcmd('failed', ['-W', work_spec['name']])
    assert manager.stdout.getvalue() == ''

def test_failed_one(manager, worker, loaded, work_spec):
    unit = manager.task_master.get_work(worker.worker_id, available_gb=16)
    assert unit.work_spec_name == work_spec['name']
    unit.fail()

    manager.runcmd('failed', ['-W', work_spec['name']])
    assert manager.stdout.getvalue() == unit.key + '\n'
    
def test_failed_one_details(manager, worker, loaded, work_spec):
    unit = manager.task_master.get_work(worker.worker_id, available_gb=16)
    assert unit.work_spec_name == work_spec['name']
    unit.fail()

    manager.runcmd('failed', ['-W', work_spec['name'], '--details'])
    assert manager.stdout.getvalue().startswith("u'" + unit.key + "': {")

def test_retry_nothing_nothing(manager, work_spec):
    manager.runcmd('retry', ['-W', work_spec['name']])
    # This case never tests whether the work spec exists
    #assert (manager.stdout.getvalue() ==
    #        'Invalid work spec \'' + work_spec['name'] + '\'.\n')
    assert manager.stdout.getvalue() == 'Nothing to do.\n'

def test_retry_a_nothing_nothing(manager, work_spec):
    manager.runcmd('retry', ['-W', work_spec['name'], '-a'])
    # This case gets an empty list for failed work units and runs silently
    #assert (manager.stdout.getvalue() ==
    #        'Invalid work spec \'' + work_spec['name'] + '\'.\n')
    assert manager.stdout.getvalue() == 'Nothing to do.\n'

def test_retry_not_loaded(manager, work_spec):
    manager.runcmd('retry', ['-W', work_spec['name'], 'key-0'])
    assert (manager.stdout.getvalue() ==
            'Invalid work spec \'' + work_spec['name'] + '\'.\n')

def test_retry_nothing(manager, loaded, work_spec):
    manager.runcmd('retry', ['-W', work_spec['name']])
    assert manager.stdout.getvalue() == 'Nothing to do.\n'

def test_retry_all_nothing(manager, loaded, work_spec):
    manager.runcmd('retry', ['-W', work_spec['name'], '-a'])
    assert manager.stdout.getvalue() == 'Nothing to do.\n'

def test_retry_missing_nothing(manager, loaded, work_spec):
    manager.runcmd('retry', ['-W', work_spec['name'], 'low-key'])
    assert manager.stdout.getvalue() == "No such failed work unit 'low-key'.\n"

def test_retry_fail_one_name(manager, worker, task_master, loaded,
                             work_spec, work_units):
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == len(work_units)

    unit = manager.task_master.get_work(worker.worker_id, available_gb=16)
    assert unit.work_spec_name == work_spec['name']
    unit_name = unit.key
    unit.fail()

    assert task_master.num_failed(work_spec['name']) == 1
    assert task_master.num_available(work_spec['name']) == len(work_units)-1

    manager.runcmd('retry', ['-W', work_spec['name'], unit_name])
    assert manager.stdout.getvalue() == 'Retried 1 work unit.\n'
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == len(work_units)

def test_retry_fail_one_all(manager, worker, task_master, loaded,
                            work_spec, work_units):
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == len(work_units)

    unit = manager.task_master.get_work(worker.worker_id, available_gb=16)
    assert unit.work_spec_name == work_spec['name']
    unit.fail()

    assert task_master.num_failed(work_spec['name']) == 1
    assert task_master.num_available(work_spec['name']) == len(work_units)-1

    manager.runcmd('retry', ['-W', work_spec['name'], '-a'])
    assert manager.stdout.getvalue() == 'Retried 1 work unit.\n'
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == len(work_units)

def test_clear_simple(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    manager.runcmd('clear', ['-W', work_spec['name']])
    assert manager.task_master.num_available(work_spec['name']) == 0
    assert (manager.stdout.getvalue() ==
            'Removed {} work units.\n'.format(len(work_units)))
    
def test_clear_available(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'available'])
    assert manager.task_master.num_available(work_spec['name']) == 0
    assert (manager.stdout.getvalue() ==
            'Removed {} work units.\n'.format(len(work_units)))

def test_clear_pending(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'pending'])
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')
    
def test_clear_blocked(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'blocked'])
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')
    
def test_clear_failed(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'failed'])
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')

def test_clear_finished(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'finished'])
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')    

def test_clear_by_name(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    a_key = work_units.keys()[0]
    cmd = ['-W', work_spec['name'], a_key]
    manager.runcmd('clear', cmd)
    logger.info('cmd %r', cmd)
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units) - 1)
    assert (manager.stdout.getvalue() == 'Removed 1 work units.\n')
    
def test_clear_available_by_name(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    a_key = work_units.keys()[0]
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'available', a_key])
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units) - 1)
    assert (manager.stdout.getvalue() == 'Removed 1 work units.\n')
    
@pytest.mark.xfail
def test_clear_pending_by_name(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    a_key = work_units.keys()[0]
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'pending', a_key])
    ### "clear -W foo -s pending bar" is known to delete an available unit
    ### named "bar"
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')
    
def test_clear_blocked_by_name(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    a_key = work_units.keys()[0]
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'blocked', a_key])
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')
    
def test_clear_failed_by_name(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    a_key = work_units.keys()[0]
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'failed', a_key])
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')
    
def test_clear_finished_by_name(manager, loaded, work_spec, work_units):
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    a_key = work_units.keys()[0]
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'finished', a_key])
    assert (manager.task_master.num_available(work_spec['name']) ==
            len(work_units))
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')    

def test_clean_with_work(manager, worked, work_spec, work_units):
    manager.runcmd('clear', ['-W', work_spec['name']])
    assert (manager.stdout.getvalue() ==
            'Removed {} work units.\n'.format(len(work_units)))
    assert manager.task_master.status(work_spec['name']) == {
        'num_available': 0,
        'num_pending': 0,
        'num_blocked': 0,
        'num_failed': 0,
        'num_finished': 0,
        'num_tasks': 0,
    }

def test_clean_available_with_work(manager, worked, work_spec, work_units):
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'available'])
    assert (manager.stdout.getvalue() ==
            'Removed {} work units.\n'.format(len(work_units) - 6))
    assert manager.task_master.status(work_spec['name']) == {
        'num_available': 0,
        'num_pending': 1,
        'num_blocked': 0,
        'num_failed': 2,
        'num_finished': 3,
        'num_tasks': 6,
    }

def test_clean_pending_with_work(manager, worked, work_spec, work_units):
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'pending'])
    assert (manager.stdout.getvalue() == 'Removed 1 work units.\n')
    assert manager.task_master.status(work_spec['name']) == {
        'num_available': len(work_units) - 6,
        'num_pending': 0,
        'num_blocked': 0,
        'num_failed': 2,
        'num_finished': 3,
        'num_tasks': len(work_units) - 1,
    }

def test_clean_blocked_with_work(manager, worked, work_spec, work_units):
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'blocked'])
    assert (manager.stdout.getvalue() == 'Removed 0 work units.\n')
    assert manager.task_master.status(work_spec['name']) == {
        'num_available': len(work_units) - 6,
        'num_pending': 1,
        'num_blocked': 0,
        'num_failed': 2,
        'num_finished': 3,
        'num_tasks': len(work_units),
    }

def test_clean_failed_with_work(manager, worked, work_spec, work_units):
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'failed'])
    assert (manager.stdout.getvalue() == 'Removed 2 work units.\n')
    assert manager.task_master.status(work_spec['name']) == {
        'num_available': len(work_units) - 6,
        'num_pending': 1,
        'num_blocked': 0,
        'num_failed': 0,
        'num_finished': 3,
        'num_tasks': len(work_units) - 2,
    }

def test_clean_finished_with_work(manager, worked, work_spec, work_units):
    manager.runcmd('clear', ['-W', work_spec['name'], '-s', 'finished'])
    assert (manager.stdout.getvalue() == 'Removed 3 work units.\n')
    assert manager.task_master.status(work_spec['name']) == {
        'num_available': len(work_units) - 6,
        'num_pending': 1,
        'num_blocked': 0,
        'num_failed': 2,
        'num_finished': 0,
        'num_tasks': len(work_units) - 3,
    }

def test_mode(manager):
    def mode(args, response):
        manager.runcmd('mode', args)
        assert manager.stdout.getvalue() == response + '\n'
        manager.stdout.truncate(0)

    mode([], "IDLE")
    mode(['run'], "set mode to 'run'")
    mode([], "RUN")
    mode(['terminate'], "set mode to 'terminate'")
    mode([], "TERMINATE")
    mode(['idle'], "set mode to 'idle'")
    mode([], "IDLE")
    
    with pytest.raises(SystemExit):
        manager.runcmd('mode', ['other'])

def test_global_lock_unlocked(manager):
    manager.runcmd('global_lock', [])
    assert manager.stdout.getvalue() == '(unlocked)\n'

def test_global_lock_locked(manager, worker):
    with worker.task_master.registry.lock(identifier=worker.worker_id):
        manager.runcmd('global_lock', [])
    assert manager.stdout.getvalue().startswith(worker.worker_id)

def test_global_lock_purge(manager, worker):
    registry = worker.task_master.registry
    with registry.lock(identifier=worker.worker_id) as session:
        manager.runcmd('global_lock', ['--purge'])
        assert manager.stdout.getvalue() == ''
        with pytest.raises(rejester.exceptions.LockError):
            registry.filter('x')

def test_workers_trivial(manager):
    manager.runcmd('workers', [])
    assert manager.stdout.getvalue() == ''

def test_workers_one(manager, worker):
    manager.runcmd('workers', [])
    assert manager.stdout.getvalue() == worker.worker_id + ' (IDLE)\n'

def test_workers_one_all(manager, worker):
    manager.runcmd('workers', ['--all']) # not very interesting
    assert manager.stdout.getvalue() == worker.worker_id + ' (IDLE)\n'

def test_workers_one_details(manager, worker):
    manager.runcmd('workers', ['--details'])
    lines = manager.stdout.getvalue().strip().split('\n')
    assert lines[0] == worker.worker_id + ' (IDLE)'
    assert len(lines) > 1
    for l in lines[1:]:
        assert l.startswith('  ')

def test_run_one_none(manager):
    manager.runcmd('run_one', [])
    assert manager.stdout.getvalue() == ''
    assert manager.exitcode == 2

def test_run_one_one(manager, work_spec, task_master, loaded):
    work_spec_name = work_spec['name']
    logger.info('work_spec name %r', work_spec_name)
    available = task_master.num_available(work_spec_name)
    logger.info('%s available: %s', work_spec_name, available)
    manager.runcmd('run_one', [])
    assert manager.stdout.getvalue() == ''
    assert manager.exitcode == 0
    assert task_master.num_available(work_spec_name) == available - 1
    assert task_master.num_pending(work_spec_name) == 0
    assert task_master.num_failed(work_spec_name) == 0
    assert task_master.num_finished(work_spec_name) == 1

def test_run_worker_args(manager, tmpdir, global_config):
    cfgfile = str(tmpdir.join('config.yaml'))
    with open(cfgfile, 'w') as f:
        f.write(yaml.dump(global_config))        
    
    for flag, value in [('--pidfile', 'foo'),  # not absolute
                        ('--logpath', 'foo'),
                        ('--pidfile', # missing dir
                         str(tmpdir.join('missing', 'pidfile'))),
                        ('--logpath', # missing dir
                         str(tmpdir.join('missing', 'logpath'))),
                        ]:
        if 'pid' in flag:
            pidfile = value
            logfile = ''
        if 'log' in flag:
            logfile = value
            pidile = ''
        rc = subprocess.call([sys.executable,
                          '-m', 'rejester.run_multi_worker', '-c', cfgfile,
                          '--pidfile', pidfile, '--logpath', logfile])
        assert rc != 0

def test_run_worker_minimal(manager, tmpdir, global_config):
    # The *only* reliable feedback we can get here is via pidfile.
    # Also, we *must* run this in a subprocess or we'll fail.
    # And, finally, we can't just fork() because of logging concerns.
    pidfile = str(tmpdir.join('pid'))
    logfile = str(tmpdir.join('log'))
    cfgfile = str(tmpdir.join('config.yaml'))
    global_config['rejester']['fork_worker'] = {
        'heartbeat_interval': 1,
    }
    with open(cfgfile, 'w') as f:
        f.write(yaml.dump(global_config))        
    
    rc = subprocess.call([sys.executable,
                          '-m', 'rejester.run_multi_worker', '-c', cfgfile,
                          '--pidfile', pidfile, '--logpath', logfile])
    assert rc == 0

    pid = None
    for i in xrange(100):
        if os.path.exists(pidfile):
            with open(pidfile, 'r') as f:
                pid = int(f.read())
            break
        time.sleep(0.1)
    assert pid, "pid file never appeared"

    # If this succeeds, the process exists
    os.kill(pid, 0)

    # Attempt a clean shutdown
    try:
        manager.runcmd('mode', ['terminate'])
        for i in xrange(600): # checks every 1-5 seconds
            try:
                os.kill(pid, 0)
            except OSError, e:
                if e.errno == errno.ESRCH: # no such process
                    pid = None
                    break
                raise
            time.sleep(0.1)
        assert pid is None, "worker failed to stop"
    finally:
        if pid is not None:
            os.kill(pid, signal.SIGKILL) # 'kill -9 pid'

def test_run_worker_sigterm(tmpdir, global_config):
    '''same as test_run_worker_minimal but shut down with a signal'''
    # The *only* reliable feedback we can get here is via pidfile.
    # Also, we *must* run this in a subprocess or we'll fail.
    # And, finally, we can't just fork() because of logging concerns.
    pidfile = str(tmpdir.join('pid'))
    logfile = str(tmpdir.join('log'))
    cfgfile = str(tmpdir.join('config.yaml'))
    with open(cfgfile, 'w') as f:
        f.write(yaml.dump(global_config))        
    
    rc = subprocess.call([sys.executable,
                          '-m', 'rejester.run_multi_worker', '-c', cfgfile,
                          '--pidfile', pidfile, '--logpath', logfile])
    assert rc == 0

    pid = None
    for i in xrange(100):
        if os.path.exists(pidfile):
            with open(pidfile, 'r') as f:
                pid = int(f.read())
            break
        time.sleep(0.1)
    assert pid, "pid file never appeared"

    # If this succeeds, the process exists
    os.kill(pid, 0)

    # Shut down with a signal
    try:
        os.kill(pid, signal.SIGTERM) # 'kill pid'
        # This really seems like it should die nearly instantly, but
        # in practice it seems to block on...something
        for i in xrange(100):
            try:
                os.kill(pid, 0)
            except OSError, e:
                if e.errno == errno.ESRCH: # no such process
                    pid = None
                    break
                raise
            time.sleep(0.1)
        assert pid is None, "worker failed to stop"
    finally:
        if pid is not None:
            os.kill(pid, signal.SIGKILL) # 'kill -9 pid'



'''Command-line :mod:`rejester_worker` tool for launching the
worker daemon.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

.. option:: --pidfile /path/to/file.pid

    If ``--pidfile`` is specified, the process ID of the worker is
    written to the named file, which must be an absolute path.

.. option:: --logpath /path/to/file.log

    If ``--logpath`` is specified, log messages from the worker will
    be written to the specified file, which again must be an absolute
    path; this is in addition to any logging specified in the
    configuration file.

 Start a worker as a background task.  The worker may be shut down by
 globally switching to ``mode terminate``, or by ``kill $(cat
 /path/to/file.pid)``.

'''
from __future__ import absolute_import
import argparse
import lockfile
import logging
import os
import sys
import time
import traceback

import daemon

import dblogger
import rejester
from rejester.exceptions import NoSuchWorkSpecError, NoSuchWorkUnitError
from rejester.run import absolute_path
from rejester.workers import run_worker, MultiWorker, SingleWorker, \
    ForkWorker
import yakonfig

logger = logging.getLogger(__name__)

def args_run_worker(parser):
    parser.add_argument('--pidfile', metavar='FILE', type=absolute_path,
                        help='file to hold process ID of worker')
    parser.add_argument('--logpath', metavar='FILE', type=absolute_path,
                        help='file to receive local logs')
    parser.add_argument('--foreground', action='store_true',
                        help='do not run the worker as a daemon')

def start_logging(gconfig, logpath):
    '''Turn on logging and set up the global config.

    This expects the :mod:`yakonfig` global configuration to be unset,
    and establishes it.  It starts the log system via the :mod:`dblogger`
    setup.  In addition to :mod:`dblogger`'s defaults, if `logpath` is
    provided, a :class:`logging.handlers.RotatingFileHandler` is set to
    write log messages to that file.

    This should not be called if the target worker is
    :class:`rejester.workers.ForkWorker`, since that manages logging on
    its own.

    :param dict gconfig: the :mod:`yakonfig` global configuration
    :param str logpath: optional location to write a log file

    '''
    yakonfig.set_default_config([rejester, dblogger], config=gconfig)
    if logpath:
        formatter = dblogger.FixedWidthFormatter()
        # TODO: do we want byte-size RotatingFileHandler or TimedRotatingFileHandler?
        handler = logging.handlers.RotatingFileHandler(
            logpath, maxBytes=10000000, backupCount=3)
        handler.setFormatter(formatter)
        logging.getLogger('').addHandler(handler)

def start_worker(which_worker, config={}):
    '''Start some worker class.

    :param str which_worker: name of the worker
    :param dict config: ``rejester`` config block

    '''
    if which_worker == 'multi_worker':
        cls = MultiWorker
    elif which_worker == 'fork_worker':
        cls = ForkWorker
    else:
        # Don't complain too hard, just fall back
        cls = ForkWorker
    return run_worker(cls, config)

def go(gconfig, args):
    '''Actually run the worker.

    This does some required housekeeping, like setting up logging for
    :class:`~rejester.workers.MultiWorker` and establishing the global
    :mod:`yakonfig` configuration.  This expects to be called with the
    :mod:`yakonfig` configuration unset.

    :param dict gconfig: the :mod:`yakonfig` global configuration
    :param args: command-line arguments

    '''
    rconfig = gconfig['rejester']
    which_worker = rconfig.get('worker', 'fork_worker')
    if which_worker == 'fork_worker':
        yakonfig.set_default_config([rejester], config=gconfig)
    else:
        start_logging(gconfig, args.logpath)
    return start_worker(which_worker, rconfig)

def fork_worker(gconfig, args):
    '''Run the worker as a daemon process.

    This uses :mod:`daemon` to run the standard double-fork, so it can
    return immediately and successfully in the parent process having forked.

    :param dict gconfig: the :mod:`yakonfig` global configuration
    :param args: command-line arguments

    '''
    if args.pidfile:
        pidfile_lock = lockfile.FileLock(args.pidfile)
    else:
        pidfile_lock = None
    with daemon.DaemonContext(pidfile=pidfile_lock, detach_process=True):
        try:
            if args.pidfile:
                with open(args.pidfile, 'w') as f:
                    f.write(str(os.getpid()))
            go(gconfig, args)
        except Exception, exc:
            logp = logpath or os.path.join('/tmp', 'rejester-failure.log')
            open(logp, 'a').write(traceback.format_exc(exc))
            raise

def main():
    parser = argparse.ArgumentParser(
        description='run a rejester work-handling daemon')
    args_run_worker(parser)
    args = yakonfig.parse_args(parser, [yakonfig, rejester])
    gconfig = yakonfig.get_global_config()
    yakonfig.clear_global_config()
    if args.foreground:
        go(gconfig, args)
    else:
        fork_worker(gconfig, args)


if __name__ == '__main__':
    main()



'''
.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''
from __future__ import absolute_import
import contextlib
import os
import copy
import time
import logging
import signal
import subprocess
import sys

import pytest
import yaml

import rejester
from rejester.workers import SingleWorker, ForkWorker
from rejester.tests.fixtures import task_master, _rejester_config  # noqa

logger = logging.getLogger(__name__)

# A note on this batch of tests: we need to talk to Redis (which we can
# do fairly synchronously), a child process, and its child processes,
# and the synchronization between all of these is hard.  Basically all
# of these are "slow" tests that depend on work units sleeping, and
# the tests being able to step in at specific times.
#
# If none of these tests run, add --runslow to the py.test command line.


def work_program(work_unit):
    # just to show that this works, we get the config from the data
    # and *reconnect* to the registry with a second instances instead
    # of using work_unit.registry
    config = work_unit.data['config']
    sleeptime = float(work_unit.data.get('sleep', 9.0))
    rejester.TaskMaster(config)  # this constructor does the reconnect
    logger.info('executing work_unit %r ... %s', work_unit.key, sleeptime)
    time.sleep(sleeptime)  # pretend to work


def work_program_broken(work_unit):
    logger.info('executing "broken" work_unit %r', work_unit.key)
    config = work_unit.data['config']
    rejester.TaskMaster(config)
    raise Exception('simulate broken work_unit')


# 1KB sized work_spec config
work_spec = dict(
    name='tbundle',
    desc='a test work bundle',
    min_gb=0.01,
    config=dict(many=' ' * 2**10, params=''),
    module='rejester.tests.test_workers',
    run_function='work_program',
    terminate_function='work_program',
)


@pytest.mark.slow  # noqa
def test_single_worker(task_master):
    work_units = {'key{}'.format(x): {'config': task_master.registry.config,
                                      'sleep': 1}
                  for x in xrange(2)}
    task_master.update_bundle(work_spec, work_units)
    assert task_master.num_finished(work_spec['name']) == 0
    assert task_master.num_available(work_spec['name']) == 2

    worker = SingleWorker(task_master.config)
    worker.register()
    rc = worker.run()
    assert rc is True
    worker.unregister()
    assert task_master.num_pending(work_spec['name']) == 0
    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) == 1
    assert task_master.num_available(work_spec['name']) == 1

    worker = SingleWorker(task_master.config)
    worker.register()
    rc = worker.run()
    assert rc is True
    worker.unregister()
    assert task_master.num_finished(work_spec['name']) == 2
    assert task_master.num_available(work_spec['name']) == 0

    worker = SingleWorker(task_master.config)
    worker.register()
    rc = worker.run()
    assert rc is False
    worker.unregister()
    assert task_master.num_finished(work_spec['name']) == 2
    assert task_master.num_available(work_spec['name']) == 0


def wait_for(timeout, condition):
    start = time.time()
    end = start + timeout
    while time.time() < end:
        if condition():
            return True
        time.sleep(0.1)
    return False


def assert_started(tm, work_spec, timeout=10):
    assert wait_for(
        timeout,
        lambda: (tm.num_available(work_spec['name']) == 0 or
                 tm.num_pending(work_spec['name']) > 0)), \
        ("worker didn't start doing work " +
         repr(tm.status(work_spec['name'])))


def assert_stopped(tm, work_spec, timeout=10):
    assert wait_for(timeout,
                    lambda: tm.num_pending(work_spec['name']) == 0), \
        ("worker didn't stop doing work " +
         repr(tm.status(work_spec['name'])))


def assert_completion(tm, work_spec, timeout=60):
    assert wait_for(
        timeout,
        lambda: (tm.num_available(work_spec['name']) == 0 and
                 tm.num_pending(work_spec['name']) == 0)), \
        ("worker didn't finish all of the work " +
         repr(tm.status(work_spec['name'])))


@contextlib.contextmanager
def run_worker(config, tmpdir, name='worker'):
    """Spawn a worker subprocess and clean up from it sanely.

    Use this as:

    >>> with run_fork_worker(task_master.registry.config, tmpdir):
    ...   while time.time() < deadline:
    ...     # do some work
    ...     time.sleep(1)

    On exiting the context block, this will do a very clean shutdown
    of the workers; if anything goes wrong this will do an unclean shutdown
    of the workers.

    """
    # Write out the config
    config_fn = tmpdir.join(name + '.yaml')
    config_fn.write(yaml.dump({'rejester': config}))

    # Start the child, with a target pid file
    #
    # (Oddly, if this test runs in the base rejester directory, the
    # re-import of this module to run the work unit fails; but explicitly
    # running the subprocess in tmpdir addresses this.)
    pid_fn = tmpdir.join(name + '.pid')
    subprocess.check_call([sys.executable,
                           '-m', 'rejester.run_multi_worker',
                           '-c', str(config_fn),
                           '--pidfile', str(pid_fn)],
                          cwd=str(tmpdir))

    # Wait for the worker to start up (up to 60s)
    assert wait_for(60, pid_fn.exists), "worker didn't start up"
    pid = int(pid_fn.read())

    # Now we're ready to go
    yield pid

    # Back?  Kill the worker
    if not ForkWorker.pid_is_alive(pid):
        return

    logger.info('stopping worker process pid={0}'.format(pid))
    os.kill(pid, signal.SIGTERM)
    if wait_for(10, lambda: not ForkWorker.pid_is_alive(pid)):
        return

    logger.info('killing worker process pid={0}'.format(pid))
    os.kill(pid, signal.SIGKILL)
    if wait_for(10, lambda: not ForkWorker.pid_is_alive(pid)):
        return

    logger.warn('worker failed to die, look for zombie pid={0}'.format(pid))


@pytest.mark.slow  # noqa
def test_worker_mode(task_master, tmpdir):
    '''Test that the worker shuts down when the mode becomes TERMINATE'''
    task_master.set_mode(task_master.RUN)
    c = dict(task_master.registry.config)
    c.setdefault('fork_worker', {})
    c['fork_worker']['heartbeat_interval'] = 1
    with run_worker(c, tmpdir) as pid:
        task_master.set_mode(task_master.TERMINATE)
        assert wait_for(5, lambda: not ForkWorker.pid_is_alive(pid)), \
            "worker failed to stop after TERMINATE"


@pytest.mark.slow  # noqa
def test_worker_idle(task_master, tmpdir):
    '''Test that the worker correctly honors most modes'''
    task_master.set_mode(task_master.IDLE)
    c = dict(task_master.registry.config)
    c.setdefault('fork_worker', {})
    c['fork_worker']['num_workers'] = 1
    c['fork_worker']['heartbeat_interval'] = 1
    c['fork_worker']['child_lifetime'] = 1
    task_master.update_bundle(work_spec, {'k1': {'config': c, 'sleep': 2},
                                          'k2': {'config': c, 'sleep': 2}})
    with run_worker(c, tmpdir):
        time.sleep(1)

        # Neither job should be going
        assert task_master.num_available(work_spec['name']) == 2

        # Let it go
        task_master.set_mode(task_master.RUN)
        assert_started(task_master, work_spec)

        assert task_master.num_available(work_spec['name']) == 1
        assert task_master.num_pending(work_spec['name']) == 1

        # Switch back to idle
        task_master.set_mode(task_master.IDLE)
        assert_stopped(task_master, work_spec)

        assert task_master.num_available(work_spec['name']) == 1
        assert task_master.num_finished(work_spec['name']) == 1

        # test_worker_mode tests the TERMINATE state


@pytest.mark.slow  # noqa
def test_do_work(task_master, tmpdir):
    num_units = 10
    num_units_cursor = 0
    c = dict(task_master.registry.config)
    work_units = {'key' + str(x): {'config': c, 'sleep': 2}
                  for x in xrange(num_units_cursor,
                                  num_units_cursor + num_units)}
    num_units_cursor += num_units
    task_master.update_bundle(work_spec, work_units)
    task_master.set_mode(task_master.RUN)

    with run_worker(c, tmpdir):
        assert_completion(task_master, work_spec)

    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) == num_units


@pytest.mark.slow  # noqa
def test_failed_work(task_master, tmpdir):
    work_spec_broken = copy.deepcopy(work_spec)
    work_spec_broken['run_function'] = 'work_program_broken'

    num_units = 10
    num_units_cursor = 0
    c = dict(task_master.registry.config)
    work_units = {'key' + str(x): {'config': c, 'sleep': 2}
                  for x in xrange(num_units_cursor,
                                  num_units_cursor + num_units)}
    num_units_cursor += num_units
    task_master.update_bundle(work_spec_broken, work_units)
    task_master.set_mode(task_master.RUN)

    with run_worker(c, tmpdir):
        assert_completion(task_master, work_spec_broken)

    assert task_master.num_failed(work_spec_broken['name']) == num_units
    assert task_master.num_finished(work_spec_broken['name']) == 0


@pytest.mark.slow  # noqa
def test_add_more_work(task_master, tmpdir):
    num_units = 10
    num_units_cursor = 0
    c = dict(task_master.registry.config)
    work_units = {'key' + str(x): {'config': c, 'sleep': 2}
                  for x in xrange(num_units_cursor,
                                  num_units_cursor + num_units)}
    num_units_cursor += num_units
    task_master.update_bundle(work_spec, work_units)
    task_master.set_mode(task_master.RUN)

    with run_worker(c, tmpdir):
        assert_completion(task_master, work_spec)

        work_units = {'key' + str(x): {'config': c, 'sleep': 2}
                      for x in xrange(num_units_cursor,
                                      num_units_cursor + num_units)}
        num_units_cursor += num_units
        task_master.update_bundle(work_spec, work_units)

        assert_completion(task_master, work_spec)

    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) == 2 * num_units


@pytest.mark.slow  # noqa
def test_add_work_midway(task_master, tmpdir):
    num_units = 10
    num_units_cursor = 0
    c = dict(task_master.registry.config)
    work_units = {'key' + str(x): {'config': c, 'sleep': 2}
                  for x in xrange(num_units_cursor,
                                  num_units_cursor + num_units)}
    num_units_cursor += num_units
    task_master.update_bundle(work_spec, work_units)
    task_master.set_mode(task_master.RUN)

    with run_worker(c, tmpdir):
        assert_started(task_master, work_spec)

        work_units = {'key' + str(x): {'config': c, 'sleep': 2}
                      for x in xrange(num_units_cursor,
                                      num_units_cursor + num_units)}
        num_units_cursor += num_units
        task_master.update_bundle(work_spec, work_units)

        assert_completion(task_master, work_spec)

    assert task_master.num_failed(work_spec['name']) == 0
    assert task_master.num_finished(work_spec['name']) == 2 * num_units


@pytest.mark.slow  # noqa
def test_fork_worker_expiry_kill(task_master, tmpdir):
    '''Test that job expiry stops a job.'''
    c = dict(task_master.registry.config)
    c['default_lifetime'] = 3  # so the job will expire quickly
    c.setdefault('fork_worker', {})
    c['fork_worker']['num_workers'] = 1
    c['fork_worker']['heartbeat_interval'] = 1
    c['fork_worker']['stop_jobs_early'] = 1

    # Create a job that sleeps for 10 seconds.  What should actually
    # happen is that, after 2-3 seconds, the "stop jobs early" monitor
    # kills it off.
    work_units = {'key0': {'config': c, 'sleep': 10}}
    task_master.update_bundle(work_spec, work_units)
    task_master.set_mode(task_master.RUN)

    with run_worker(c, tmpdir):
        assert_started(task_master, work_spec)
        time.sleep(5.0)
        # Now (some synchronization issues aside) the job *should*
        # be dead
        assert task_master.status(work_spec['name']) == {
            'num_available': 0, 'num_pending': 0,
            'num_finished': 0, 'num_failed': 1,
            'num_blocked': 0, 'num_tasks': 1,
        }


@pytest.mark.slow  # noqa
def test_fork_worker_expiry_dup(task_master, tmpdir):
    '''Test that job expiration and default_lifetime work as expected.'''
    # Create a job that sleeps for 10 seconds
    work_units = {'key0': {'config': task_master.registry.config,
                           'sleep': 10.0}}
    task_master.update_bundle(work_spec, work_units)
    task_master.set_mode(task_master.RUN)

    assert task_master.status(work_spec['name']) == {
        'num_available': 1, 'num_pending': 0,
        'num_finished': 0, 'num_failed': 0,
        'num_blocked': 0, 'num_tasks': 1,
    }

    # Start a ForkWorker, but with a 1-second job expiration
    c1 = dict(task_master.registry.config)
    c1['default_lifetime'] = 1
    c1['worker'] = 'fork_worker'
    c1.setdefault('fork_worker', {})
    c1['fork_worker']['num_workers'] = 1
    with run_worker(c1, tmpdir, 'c1'):
        assert_started(task_master, work_spec)
        time.sleep(2.0)

        # Now the worker is still doing the job, but it's also in the
        # available queue again
        assert task_master.status(work_spec['name']) == {
            'num_available': 1, 'num_pending': 0,
            'num_finished': 0, 'num_failed': 0,
            'num_blocked': 0, 'num_tasks': 1,
        }

        # So start a second worker, with the default timeout
        c2 = dict(task_master.registry.config)
        c2['worker'] = 'fork_worker'
        c2.setdefault('fork_worker', {})
        c2['fork_worker']['num_workers'] = 1
        with run_worker(c2, tmpdir, 'c2'):
            assert_started(task_master, work_spec)

            # Now the second worker should have picked up the job
            assert task_master.status(work_spec['name']) == {
                'num_available': 0, 'num_pending': 1,
                'num_finished': 0, 'num_failed': 0,
                'num_blocked': 0, 'num_tasks': 1,
            }

            # Wait for it to finish
            assert_completion(task_master, work_spec)

            # Now: both workers should have finished the job; the first
            # worker catches LostLease and reports nothing; the second
            # worker reports successful completion
            assert task_master.status(work_spec['name']) == {
                'num_available': 0, 'num_pending': 0,
                'num_finished': 1, 'num_failed': 0,
                'num_blocked': 0, 'num_tasks': 1,
            }



'''

'''
from __future__ import absolute_import
from collections import defaultdict
import logging
import os
import sys
import time

import pytest

import rejester
from rejester.exceptions import EnvironmentError, LockError

logger = logging.getLogger(__name__)
pytest_plugins = 'rejester.tests.fixtures'

def test_registry_decode(registry):
    for foo in [{}, None, '', 0, 1]:
        assert registry._decode(registry._encode(foo)) == foo


def test_registry_lock_block(registry):
    with registry.lock(ltime=100) as session1:
        ## check that we cannot get the lock
        with pytest.raises(rejester.exceptions.LockError):
            with registry.lock(atime=1) as session2:
                pass


def test_registry_lock_loss(registry, registry2):
    with pytest.raises(LockError) as env_error:
        with registry.lock(ltime=1) as session1:
            ## check that we can get the lock
            with registry2.lock(atime=100) as session2:
                assert session2
            ## now try to do anything with the expired session
            session1.filter('x')


def test_registry_re_acquire_lock(registry, registry2):
    with registry.lock(ltime=1000) as session:
        ## lower the ltime
        assert session.re_acquire_lock(ltime=1)
        ## capture it with atime so short it couldn't happen if
        ## ltime were not reset
        with registry2.lock(atime=10) as session2:
            assert session2
        ## verify that re_acquire_lock fails
        with pytest.raises(EnvironmentError):
            session.re_acquire_lock()


def test_registry_update_pull(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock() as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict


def test_registry_filter(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock() as session:
        session.update('test_dict', test_dict, 
                       priorities=dict(cars=100, houses=-100))
        assert session.filter('test_dict') == test_dict
        assert session.filter('test_dict', priority_min=0) == dict(cars=10)
        assert session.filter('test_dict', priority_max=0) == dict(houses=5)


def test_registry_update_expire(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict, expire=2)
        assert session.pull('test_dict') == test_dict
        time.sleep(3)
        assert session.pull('darn') == {}
        assert session.pull('test_dict') == {}


def test_registry_get(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict
        assert session.get('test_dict', 'cars') == 10
        assert session.get('test_dict', 'not-there', 'hello') == 'hello'


def test_registry_set(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock(atime=5000) as session:
        session.set('test_dict', 'cars', 10)
        session.set('test_dict', 'houses', 5)
        assert session.pull('test_dict') == test_dict


def test_registry_set_priority(registry):
    with registry.lock(atime=5000) as session:
        session.set('test_dict', 'cars', 10, 100)
        assert session.popitem('test_dict', priority_max=0) is None
        assert session.popitem('test_dict', priority_max=200) == ('cars', 10)


def test_registry_delete(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock(atime=5000) as session:
        session.set('test_dict', 'cars', 10)
        session.set('test_dict', 'houses', 5)
        assert session.pull('test_dict') == test_dict
        session.delete('test_dict')
        assert session.pull('test_dict') == {}
        assert session.popitem('test_dict') is None


def test_registry_popmany(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict
        session.popmany('test_dict', 'dogs', 'cars', 'houses')
        assert session.pull('test_dict') == dict()


def test_registry_popitem(registry):
    test_dict = dict(cars=10, houses=5)

    recovered = set()
    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict
        recovered.add(session.popitem('test_dict'))
        recovered.add(session.popitem('test_dict'))
        assert recovered == set(test_dict.items())

        assert session.popitem('test_dict') is None


def test_registry_popitem_priority(registry):
    test_dict = dict(cars=10, houses=5, dogs=99)

    recovered = set()
    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict, priorities=defaultdict(lambda: 10))
        assert session.pull('test_dict') == test_dict
        assert session.popitem('test_dict', priority_min=100) is None
        recovered.add(session.popitem('test_dict', priority_min=-100))
        recovered.add(session.popitem('test_dict', priority_min=10))
        recovered.add(session.popitem('test_dict'))
        assert recovered == set(test_dict.items())


def test_registry_popitem_move(registry):
    test_dict = dict(cars=10, houses=5)

    recovered = set()
    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict

        recovered.add(session.popitem_move('test_dict', 'second'))
        assert session.len('test_dict') == 1
        assert session.len('second') == 1
        recovered.add(session.popitem_move('test_dict', 'second'))
        assert session.len('test_dict') == 0
        assert session.len('second') == 2

        assert recovered == set(test_dict.items())
        assert recovered == set(session.pull('second').items())


def test_registry_popitem_move_all(registry):
    test_dict = dict(cars=10, houses=5)

    recovered = set()
    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict, priorities=defaultdict(lambda: 100))
        session.move_all('test_dict', 'second')
        ## check that priorities were also moved
        assert session.popitem('second', priority_max=0) is None
        assert session.pull('second') == test_dict        
        assert session.len('test_dict') == 0


def test_registry_popitem_move_priority(registry):
    test_dict = dict(cars=10, houses=5, dogs=99)

    recovered = set()
    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict, priorities=defaultdict(lambda: 10))
        assert session.pull('test_dict') == test_dict
        assert session.popitem_move('test_dict','second',  priority_min=100) is None
        assert session.popitem_move('test_dict','second',  priority_max=-100) is None
        recovered.add(session.popitem_move('test_dict', 'second', priority_min=-100))
        recovered.add(session.popitem_move('test_dict', 'second', priority_min=10))
        recovered.add(session.popitem_move('test_dict', 'second'))
        assert recovered == set(test_dict.items())
        assert recovered == set(session.pull('second').items())


def test_registry_popitem_move_empty(registry):
    test_dict = dict(cars=10, houses=5)

    recovered = set()
    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict

        session.popitem_move('test_dict', 'second')
        session.popitem_move('test_dict', 'second')
        assert session.len('test_dict') == 0

        assert session.popitem_move('test_dict', 'second') is None

def test_registry_move(registry):
    test_dict = dict(cars=10, houses=5, dogs=4)

    recovered = set()
    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict

        assert session.len('test_dict') == 3
        session.move('test_dict', 'second', dict(cars=3))
        assert session.len('test_dict') == 2
        assert session.len('second') == 1
        session.move('test_dict', 'second', dict(houses=2))
        assert session.len('test_dict') == 1
        assert session.len('second') == 2

        assert dict(cars=3, houses=2) == session.pull('second')
        assert dict(dogs=4) == session.pull('test_dict')

def test_registry_getitem_reset_priorities(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock() as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict
        session.reset_priorities('test_dict', -100)
        assert session.popitem('test_dict', priority_min=0) is None
        assert session.popitem('test_dict', priority_min=-100) 
        assert session.popitem('test_dict', priority_min=-100) 

def test_registry_getitem_reset(registry):
    test_dict = dict(cars=10, houses=5)

    recovered1 = set()
    recovered2 = set()
    with registry.lock(atime=5000) as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict

        assert session.getitem_reset('test_dict', priority_min=10) is None

        recovered1.add(session.getitem_reset('test_dict', priority_min=-10, new_priority=10))
        recovered1.add(session.getitem_reset('test_dict', priority_min=-10, priority_max=8, new_priority=10))
        assert recovered1 == set(test_dict.items())

        assert session.getitem_reset('test_dict', priority_max=9) is None

        recovered2.add(session.getitem_reset('test_dict', priority_max=10, new_priority=20))
        recovered2.add(session.getitem_reset('test_dict', priority_max=10))
        assert recovered2 == set(test_dict.items())
        
    
def test_registry_increment(registry):

    with registry.lock(atime=5000) as session:
        session.increment('test_dict', 'dog')
        assert session.pull('test_dict') == dict(dog=1)

        session.increment('test_dict', 'foo', 4)
        assert session.pull('test_dict') == dict(dog=1, foo=4)
        session.increment('test_dict', 'foo', 0.5)
        assert session.pull('test_dict') == dict(dog=1, foo=4.5)
        session.increment('test_dict', 'foo', 1)
        assert session.pull('test_dict') == dict(dog=1, foo=5.5)
        session.increment('test_dict', 'foo')
        assert session.pull('test_dict') == dict(dog=1, foo=6.5)


def test_registry_1to1(registry):
    
    with registry.lock() as session:
        session.set_1to1('test_dict', 'k1', 'v1')
        assert session.get('test_dict', 'k1') == 'v1'
        assert session.get('test_dict', 'v1') == 'k1'

        session.set_1to1('test_dict', 'v1', 'k3')
        assert session.get('test_dict', 'k3') == 'v1'
        assert session.get('test_dict', 'v1') == 'k3'


def test_registry_update_locks(registry):

    test_dict = dict(dog=10, cat=42)
    test_locks = dict(dog='w1', cat='w3')

    with registry.lock() as session:
        session.update('test_dict', test_dict, locks=test_locks)
        assert session.pull('test_dict') == test_dict

        session.update('test_dict', dict(dog=4), locks=dict(dog='w1'))
        assert session.get('test_dict', 'dog') == 4

        with pytest.raises(EnvironmentError):
            session.update('test_dict', dict(dog=8), locks=dict(dog='w3'))

        assert session.get('test_dict', 'dog') == 4

def test_registry_getitem_reset_lock(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock() as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict
        k1, v1 = session.getitem_reset('test_dict', lock='w1', new_priority=100)
        k3, v3 = session.getitem_reset('test_dict', lock='w3', priority_max=100)

        logger.info({k1: v1, k3: v3})
        logger.info(session.pull('test_dict_locks'))
        good_locks = {k1: 'w1',  k3: 'w3', 'w1':  k1, 'w3': k3}
        bad_locks  = {k1: 'w1-', k3: 'w3', 'w1-': k1, 'w3': k3}
        assert good_locks == session.pull('test_dict_locks')
        assert  bad_locks != session.pull('test_dict_locks')
        logger.info('good_locks: %r, bad_locks: %r', good_locks, bad_locks)

        session.update('test_dict', mapping=test_dict, locks=good_locks)

        with pytest.raises(EnvironmentError):
            session.update('test_dict', mapping=test_dict, locks={k1: 'w1-', k3: 'w3'})

def test_registry_getitem_reset_lock_1to1(registry):
    test_dict = dict(cars=10, houses=5)

    with registry.lock() as session:
        session.update('test_dict', test_dict)
        assert session.pull('test_dict') == test_dict
        k1, v1 = session.getitem_reset('test_dict', lock='w1', new_priority=100)
        k3, v3 = session.getitem_reset('test_dict', lock='w3', priority_max=100)

        locks = session.pull('test_dict_locks')
        assert 'w1' in locks
        assert locks['w1'] == session.get('test_dict_locks', 'w1')
